from workbench import generic
from workbench.accounts.models import Team, User
from workbench.planning import reporting
from workbench.projects.models import Project


class ProjectPlanningView(generic.DetailView):
    model = Project
    template_name = "planning/project_planning.html"

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            planning_data=reporting.project_planning(self.object), **kwargs
        )


class UserPlanningView(generic.DetailView):
    model = User
    template_name = "planning/user_planning.html"

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            planning_data=reporting.user_planning(self.object), **kwargs
        )


class TeamPlanningView(generic.DetailView):
    model = Team
    template_name = "planning/team_planning.html"

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            planning_data=reporting.team_planning(self.object), **kwargs
        )



from django.test import TestCase

from workbench import factories


class RedirectsTest(TestCase):
    def test_redirects(self):
        """Root redirects work correctly"""
        self.client.force_login(factories.UserFactory.create())

        response = self.client.get("/projects/projects/42/")
        self.assertRedirects(response, "/projects/42/", fetch_redirect_response=False)

        response = self.client.get("/offers/42/")
        self.assertRedirects(
            response, "/projects/offers/42/", fetch_redirect_response=False
        )



import datetime as dt
from decimal import Decimal

from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils.translation import deactivate_all

from time_machine import travel

from workbench import factories
from workbench.audit.models import LoggedAction
from workbench.offers.models import Offer
from workbench.projects.models import Project
from workbench.tools.formats import local_date_format
from workbench.tools.forms import WarningsForm
from workbench.tools.testing import check_code, messages
from workbench.tools.validation import in_days


class OffersTest(TestCase):
    def tearDown(self):
        deactivate_all()

    def test_create_offer(self):
        """Create an offer and generate a PDF"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        service = factories.ServiceFactory.create(
            project=project, effort_type="Programming", effort_rate=200, effort_hours=10
        )

        url = project.urls["createoffer"]
        response = self.client.get(url)
        self.assertContains(response, 'id="id_postal_address"')
        self.assertNotContains(response, 'data-field-value="')
        postal_address = factories.PostalAddressFactory.create(person=project.contact)
        response = self.client.get(url)
        self.assertContains(response, 'id="id_postal_address"')
        self.assertContains(response, 'data-field-value="')

        response = self.client.post(
            url,
            {
                "title": "Stuff",
                "owned_by": project.owned_by_id,
                "discount": "10",
                "liable_to_vat": "1",
                "postal_address": postal_address.postal_address,
                "services": [service.id],
                "status": Offer.IN_PREPARATION,
            },
        )
        self.assertEqual(response.status_code, 302)

        offer = Offer.objects.get()
        self.assertRedirects(response, offer.get_absolute_url())
        self.assertAlmostEqual(offer.total_excl_tax, Decimal("1990"))
        self.assertAlmostEqual(offer.total, Decimal("2143.25"))

        pdf = self.client.get(offer.urls["pdf"])
        self.assertEqual(pdf.status_code, 200)  # No crash
        self.assertEqual(pdf["content-type"], "application/pdf")

        offer.show_service_details = True
        offer.save()

        pdf = self.client.get(offer.urls["pdf"])
        self.assertEqual(pdf.status_code, 200)  # No crash
        self.assertEqual(pdf["content-type"], "application/pdf")

        # Deleting the service automagically updates the offer
        offer.services.get().delete()
        offer.refresh_from_db()
        self.assertAlmostEqual(offer.total_excl_tax, Decimal("-10"))

        pdf = self.client.get(offer.urls["pdf"])
        self.assertEqual(pdf.status_code, 200)  # No crash
        self.assertEqual(pdf["content-type"], "application/pdf")

    def test_offers_pdf(self):
        """Generate a PDF containing several offers"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        response = self.client.get(project.urls["offers_pdf"])
        self.assertRedirects(response, project.get_absolute_url())
        self.assertEqual(messages(response), ["No offers in project."])

        offer = factories.OfferFactory.create(project=project)

        response = self.client.get(project.urls["offers_pdf"])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/pdf")

        offer.offered_on = dt.date.today()
        offer.valid_until = in_days(60)
        offer.save()
        project.description = "Test"
        project.save()

        response = self.client.get(project.urls["offers_pdf"])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/pdf")

    def test_update_offer(self):
        """Offer and bound services update warnings and errors"""
        offer = factories.OfferFactory.create(title="Test")
        service = factories.ServiceFactory.create(
            project=offer.project,
            effort_type="Programming",
            effort_rate=200,
            effort_hours=10,
        )
        self.client.force_login(offer.owned_by)

        response = self.client.get(service.urls["update"])
        self.assertRedirects(response, service.get_absolute_url())

        response = self.client.get(offer.urls["delete"])
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            offer.urls["update"],
            {
                "title": "Stuff",
                "owned_by": offer.owned_by_id,
                "discount": "10",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "services": [service.id],
                # "offered_on": dt.date.today().isoformat(),
                "status": Offer.ACCEPTED,
            },
        )
        self.assertContains(response, "Offered on date missing for selected state.")
        self.assertContains(response, "Valid until date missing for selected state.")

        response = self.client.post(
            offer.urls["update"],
            {
                "title": "Stuff",
                "owned_by": offer.owned_by_id,
                "discount": "10",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "services": [service.id],
                "offered_on": dt.date.today().isoformat(),
                "valid_until": in_days(60).isoformat(),
                "status": Offer.ACCEPTED,
            },
        )
        self.assertRedirects(response, offer.get_absolute_url())

        response = self.client.get(offer.urls["update"])
        self.assertContains(
            response,
            "This offer is not in preparation anymore. I assume you know what you",
        )

        offer.refresh_from_db()
        self.assertEqual(offer.closed_on, dt.date.today())
        self.assertAlmostEqual(offer.subtotal, Decimal("2000"))

        response = self.client.get(service.urls["detail"])
        self.assertRedirects(
            response, "%s#service%s" % (offer.project.urls["detail"], service.pk)
        )

        response = self.client.get(
            service.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response,
            "Most fields are disabled because service is bound"
            " to an offer which is not in preparation anymore.",
        )

        self.assertTrue(service.allow_logging)
        response = self.client.post(
            service.urls["update"],
            {
                "allow_logging": False,
                WarningsForm.ignore_warnings_id: "no-role-selected",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 202)
        service.refresh_from_db()
        self.assertFalse(service.allow_logging)
        self.assertEqual(
            messages(response), ["Service 'Any service' has been updated successfully."]
        )
        self.assertEqual(service.offer, offer)

        response = self.client.get(
            service.urls["delete"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertContains(
            response,
            "Cannot delete a service bound to an offer"
            " which is not in preparation anymore.",
        )

        response = self.client.post(
            offer.urls["update"],
            {
                "title": "Stuff",
                "owned_by": offer.owned_by_id,
                "discount": "10",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "services": [service.id],
                "offered_on": dt.date.today().isoformat(),
                "status": Offer.IN_PREPARATION,
            },
        )
        self.assertContains(response, "but the offer")

        response = self.client.post(
            offer.urls["update"],
            {
                "title": "Stuff",
                "owned_by": offer.owned_by_id,
                "discount": "10",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "services": [service.id],
                "offered_on": dt.date.today().isoformat(),
                "status": Offer.IN_PREPARATION,
                WarningsForm.ignore_warnings_id: "status-change-but-already-closed",
            },
        )
        self.assertRedirects(response, offer.get_absolute_url())

        offer.refresh_from_db()
        self.assertIsNone(offer.closed_on)

    def test_list(self):
        """Filter form smoke test"""
        offer = factories.OfferFactory.create()
        self.client.force_login(factories.UserFactory.create())

        code = check_code(self, Offer.urls["list"])
        code("")
        code("q=test")
        code("s=all")
        code("s=10")
        code("s=20")
        code("s=20")
        code("org={}".format(offer.project.customer_id))
        code("owned_by={}".format(offer.owned_by_id))
        code("owned_by=-1")  # mine
        code("owned_by=0")  # only inactive

    def test_detail(self):
        """The detail URL of offers redirects to the project detail page anchor"""
        offer = factories.OfferFactory.create()
        self.client.force_login(offer.owned_by)

        self.assertRedirects(
            self.client.get("{}{}/".format(Offer.urls["list"], offer.id)),
            "{}#offer{}".format(offer.project.get_absolute_url(), offer.id),
        )

    def test_create_message(self):
        """Attempting to directly create an offer only produces a help message"""
        self.client.force_login(factories.UserFactory.create())
        response = self.client.get(Offer.urls["create"])
        self.assertRedirects(response, Offer.urls["list"])
        self.assertEqual(
            messages(response),
            [
                "Offers can only be created from projects. Go to the project"
                " and add services first, then you'll be able to create the offer"
                " itself."
            ],
        )

    def test_status(self):
        """Offer status badge"""
        project = factories.ProjectFactory.create()
        today = dt.date.today()
        self.assertEqual(
            Offer(project=project, status=Offer.IN_PREPARATION).pretty_status,
            "In preparation since {}".format(local_date_format(today)),
        )
        self.assertEqual(
            Offer(
                project=project,
                status=Offer.OFFERED,
                offered_on=today,
                valid_until=today,
            ).pretty_status,
            "Offered on {}".format(local_date_format(today)),
        )
        self.assertEqual(
            Offer(
                project=project, status=Offer.DECLINED, closed_on=today
            ).pretty_status,
            "Declined on {}".format(local_date_format(today)),
        )
        self.assertEqual(Offer(project=project, status="42").pretty_status, "42")

    def test_declined_offer(self):
        """Declined offers' services do not allow logging and are not part of
        service hours"""
        project = factories.ProjectFactory.create()

        offer1 = factories.OfferFactory.create(project=project, status=Offer.DECLINED)
        offer2 = factories.OfferFactory.create(project=project)

        factories.ServiceFactory.create(project=project, offer=offer1, effort_hours=10)
        factories.ServiceFactory.create(project=project, offer=offer2, effort_hours=20)
        factories.ServiceFactory.create(project=project, effort_hours=40)

        gs = project.grouped_services
        self.assertEqual([offer2, None, offer1], [row[0] for row in gs["offers"]])
        self.assertEqual(gs["service_hours"], Decimal("60.00"))  # Not 70

        self.assertEqual(project.services.logging().count(), 2)

    def test_offer_copying(self):
        """Offers can be copied to different projects"""
        offer = factories.OfferFactory.create()
        factories.ServiceFactory.create(project=offer.project, offer=offer)

        self.client.force_login(offer.owned_by)
        response = self.client.get(offer.urls["copy"])
        self.assertContains(
            response,
            'data-autocomplete-url="{}?only_open=on"'.format(
                Project.urls["autocomplete"]
            ),
        )

        response = self.client.post(
            offer.urls["copy"], {"modal-project": offer.project.pk}
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'value="same-project"')

        project = factories.ProjectFactory.create()
        response = self.client.post(offer.urls["copy"], {"modal-project": project.pk})
        self.assertEqual(response.status_code, 299)
        offer = project.offers.first()
        self.assertEqual(
            response.json(),
            {"redirect": "%s#offer%s" % (project.get_absolute_url(), offer.pk)},
        )

        self.assertEqual(project.offers.count(), 1)
        self.assertEqual(project.services.count(), 1)

    def test_postal_address_validation(self):
        """Postal addresses are validated (a bit)"""
        project = factories.ProjectFactory.create()
        self.client.force_login(project.owned_by)

        url = project.urls["createoffer"]
        response = self.client.post(
            url,
            {
                "title": "Stuff",
                "owned_by": project.owned_by_id,
                "discount": "10",
                "liable_to_vat": "1",
                "postal_address": "Anything",
                "status": Offer.IN_PREPARATION,
            },
        )
        self.assertContains(response, 'value="short-postal-address"')

    def test_grouped_services_empty_offers(self):
        """Offers without services still appear in Project.grouped_services"""
        offer = factories.OfferFactory.create()
        gs = offer.project.grouped_services

        self.assertEqual(gs["offers"][0][0], offer)
        self.assertEqual(gs["offers"][0][1]["services"], [])

    def test_offer_deletion_without_logbook(self):
        """Offers without linked logged hours can be deleted, including all services"""
        offer = factories.OfferFactory.create()
        self.client.force_login(offer.owned_by)

        response = self.client.get(offer.urls["delete"])
        self.assertContains(response, "delete_services")

        response = self.client.post(offer.urls["delete"], {"delete_services": "on"})
        self.assertRedirects(response, offer.project.urls["detail"])

        self.assertEqual(offer.project.services.count(), 0)

    def test_offer_deletion_with_logbook(self):
        """Offers with some linked logged hours can be deleted, but the service stays"""
        offer = factories.OfferFactory.create()
        self.client.force_login(offer.owned_by)
        service = factories.ServiceFactory.create(project=offer.project, offer=offer)
        factories.LoggedHoursFactory.create(service=service)

        response = self.client.get(offer.urls["delete"])
        self.assertNotContains(response, "delete_services")

        response = self.client.post(offer.urls["delete"], {"delete_services": "on"})
        self.assertRedirects(response, offer.project.urls["detail"])

        self.assertEqual(offer.project.services.count(), 1)

    def test_offer_pricing(self):
        """Offer total calculation works"""
        offer = factories.OfferFactory.create()
        service = factories.ServiceFactory.create(offer=offer, project=offer.project)

        self.client.force_login(offer.owned_by)
        response = self.client.get(offer.urls["pricing"])
        self.assertContains(response, "data-effort-rate", 2)  # Empty form too

        response = self.client.post(
            offer.urls["pricing"],
            {
                "discount": 100,
                "liable_to_vat": "on",
                "services-TOTAL_FORMS": 1,
                "services-INITIAL_FORMS": 1,
                "services-MAX_NUM_FORMS": 1000,
                "services-0-id": service.id,
                "services-0-offer": offer.id,
                "services-0-title": service.title,
                "services-0-effort_type": "Programming",
                "services-0-effort_rate": 250,
                "services-0-effort_hours": 10,
                "services-0-cost": 35,
            },
        )

        self.assertRedirects(response, offer.get_absolute_url())

        offer.refresh_from_db()
        service.refresh_from_db()

        self.assertEqual(service.service_cost, 250 * 10 + 35)
        self.assertEqual(offer.total_excl_tax, 250 * 10 + 35 - 100)

        # Offer has been
        # - inserted
        # - updated once because of _fts
        # - updated only once because of skip_related_model
        actions = LoggedAction.objects.for_model(offer).with_data(id=offer.id)
        self.assertEqual([action.action for action in actions], ["I", "U", "U"])

    def test_offer_pricing_with_flat_rate(self):
        """Pricing with flat rates does not allow editing effort rates"""
        project = factories.ProjectFactory.create(flat_rate=160)
        offer = factories.OfferFactory.create(project=project)
        factories.ServiceFactory.create(offer=offer, project=project)

        self.client.force_login(project.owned_by)
        response = self.client.get(offer.urls["pricing"])
        self.assertContains(
            response,
            '<input type="number" name="services-0-effort_rate" step="0.01" class="form-control" disabled id="id_services-0-effort_rate">',  # noqa
            html=True,
        )

    def test_renumber_offers(self):
        """Offers can be renumbered (and reordered)"""
        project = factories.ProjectFactory.create()
        offer1 = factories.OfferFactory.create(project=project)
        offer2 = factories.OfferFactory.create(project=project)
        field1 = "offer_{}_code".format(offer1.pk)
        field2 = "offer_{}_code".format(offer2.pk)

        self.client.force_login(project.owned_by)
        response = self.client.get(project.urls["renumber_offers"])
        self.assertContains(response, field1)
        self.assertContains(response, field2)

        response = self.client.post(
            project.urls["renumber_offers"], {field1: 1, field2: 1}
        )
        self.assertContains(response, "Codes must be unique")

        response = self.client.post(
            project.urls["renumber_offers"], {field1: 4, field2: 3}
        )
        self.assertRedirects(response, project.urls["detail"])

        offer1.refresh_from_db()
        offer2.refresh_from_db()

        self.assertEqual(offer1._code, 4)
        self.assertEqual(offer2._code, 3)

    def test_please_decline(self):
        """Please do not decline offers but update them and send them again"""
        offer = factories.OfferFactory.create()
        self.client.force_login(offer.owned_by)

        response = self.client.post(
            offer.urls["update"],
            {
                "title": "Stuff",
                "owned_by": offer.owned_by_id,
                "discount": "10",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "offered_on": in_days(0).isoformat(),
                "valid_until": in_days(60).isoformat(),
                "status": Offer.DECLINED,
            },
        )
        self.assertContains(
            response,
            "However, if you just want to change a few things and send the offer",
        )

        response = self.client.post(
            offer.urls["update"],
            {
                "title": "Stuff",
                "owned_by": offer.owned_by_id,
                "discount": "10",
                "liable_to_vat": "1",
                "postal_address": "Anything\nStreet\nCity",
                "offered_on": in_days(0).isoformat(),
                "valid_until": in_days(60).isoformat(),
                "status": Offer.DECLINED,
                WarningsForm.ignore_warnings_id: "yes-please-decline",
            },
        )
        self.assertRedirects(
            response, offer.get_absolute_url(), fetch_redirect_response=False
        )

        offer.refresh_from_db()  # Refresh the offer title etc.
        self.assertEqual(
            messages(response),
            [
                "All offers of project {} are declined. You might "
                "want to close the project now?".format(offer.project),
                "Offer '{}' has been updated successfully.".format(offer),
            ],
        )

        offer.project.closed_on = dt.date.today()
        self.assertIsNone(offer.project.solely_declined_offers_warning(request=None))

    @travel("2020-04-21 12:00")
    def test_offer_with_closed_project(self):
        """Sent offers on closed projects have a warning badge"""
        project = factories.ProjectFactory.create(closed_on=dt.date.today())
        offer = factories.OfferFactory.create(
            project=project, offered_on=dt.date.today(), status=Offer.OFFERED
        )

        self.assertEqual(
            offer.status_badge,
            '<span class="badge badge-warning">Offered on 21.04.2020, but project closed on 21.04.2020</span>',  # noqa
        )

    @travel("2020-05-04 12:00")
    def test_offer_not_valid_anymore(self):
        """Sent offers which are not valid anymore have a warning badge"""
        project = factories.ProjectFactory.create()
        offer = factories.OfferFactory.create(
            project=project,
            offered_on=in_days(-90),
            valid_until=in_days(-30),
            status=Offer.OFFERED,
        )

        self.assertEqual(
            offer.status_badge,
            '<span class="badge badge-warning">Offered on 04.02.2020, but not valid anymore</span>',  # noqa
        )

    def test_offers_ordering(self):
        """The ordering of offers is: Valid offers, not offered yet, declined offers"""
        project = factories.ProjectFactory.create()

        declined = factories.OfferFactory.create(
            project=project, offered_on=dt.date.today(), status=Offer.DECLINED
        )
        accepted = factories.OfferFactory.create(
            project=project, offered_on=dt.date.today(), status=Offer.ACCEPTED
        )
        in_preparation = factories.OfferFactory.create(project=project)
        factories.ServiceFactory.create(project=project)

        self.assertTrue(accepted < in_preparation)  # _code
        self.assertTrue(in_preparation < None)
        self.assertTrue(None < declined)
        self.assertTrue(in_preparation < declined)

        self.assertEqual(
            [row[0] for row in project.grouped_services["offers"]],
            [accepted, in_preparation, None, declined],
        )
        self.assertEqual(in_preparation < 42, 0)

    def test_model_validation(self):
        """Offer model validation"""
        offer = Offer(
            title="Test",
            project=factories.ProjectFactory.create(),
            owned_by=factories.UserFactory.create(),
            status=Offer.OFFERED,
            postal_address="Test\nStreet\nCity",
            _code=1,
        )

        with self.assertRaises(ValidationError) as cm:
            offer.clean_fields()
        self.assertEqual(
            list(cm.exception),
            [
                ("offered_on", ["Offered on date missing for selected state."]),
                ("valid_until", ["Valid until date missing for selected state."]),
            ],
        )

        with self.assertRaises(ValidationError) as cm:
            offer.offered_on = in_days(0)
            offer.valid_until = in_days(-1)
            offer.full_clean()
        self.assertEqual(
            list(cm.exception),
            [("valid_until", ["Valid until date has to be after offered on date."])],
        )

    def test_properties(self):
        """Offer property testing"""
        self.assertTrue(Offer(status=Offer.ACCEPTED).is_accepted)
        self.assertTrue(Offer(status=Offer.DECLINED).is_declined)



# Generated by Django 2.1.7 on 2019-03-04 21:39

from decimal import Decimal

import django.core.validators
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="Offer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "subtotal",
                    models.DecimalField(
                        decimal_places=2,
                        default=Decimal("0.00"),
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="subtotal",
                    ),
                ),
                (
                    "discount",
                    models.DecimalField(
                        decimal_places=2,
                        default=Decimal("0.00"),
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="discount",
                    ),
                ),
                (
                    "liable_to_vat",
                    models.BooleanField(
                        default=True,
                        help_text="For example invoices to foreign institutions are not liable to VAT.",
                        verbose_name="liable to VAT",
                    ),
                ),
                (
                    "tax_rate",
                    models.DecimalField(
                        decimal_places=2,
                        default=Decimal("7.7"),
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="tax rate",
                    ),
                ),
                (
                    "total",
                    models.DecimalField(
                        decimal_places=2,
                        default=Decimal("0.00"),
                        max_digits=10,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="total",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                (
                    "offered_on",
                    models.DateField(blank=True, null=True, verbose_name="offered on"),
                ),
                (
                    "closed_on",
                    models.DateField(blank=True, null=True, verbose_name="closed on"),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                (
                    "description",
                    models.TextField(blank=True, verbose_name="description"),
                ),
                (
                    "status",
                    models.PositiveIntegerField(
                        choices=[
                            (10, "In preparation"),
                            (20, "Offered"),
                            (30, "Accepted"),
                            (40, "Declined"),
                        ],
                        default=10,
                        verbose_name="status",
                    ),
                ),
                ("postal_address", models.TextField(verbose_name="postal address")),
                ("_code", models.IntegerField(verbose_name="code")),
                (
                    "owned_by",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="responsible",
                    ),
                ),
            ],
            options={
                "verbose_name": "offer",
                "verbose_name_plural": "offers",
                "ordering": ["-pk"],
            },
        )
    ]



from django.conf import settings
from django.test import Client, TestCase
from django.utils.translation import deactivate_all

from workbench import factories
from workbench.accounts import views
from workbench.accounts.models import User
from workbench.tools.testing import messages


class FakeFlow:
    EMAIL = "user@example.com"
    RAISE_EXCEPTION = False

    def get_authentication_url(self):
        return "http://example.com/auth/"

    def get_user_data(self):
        if self.RAISE_EXCEPTION:
            raise Exception("Email not verified")
        return {"email": self.EMAIL}


views.GoogleOAuth2Client = lambda *a, **kw: FakeFlow()


class LoginTestCase(TestCase):
    def tearDown(self):
        deactivate_all()

    def test_login(self):
        """Basic login and logout redirects"""
        user = factories.UserFactory.create()

        self.assertRedirects(self.client.get("/"), "/accounts/login/")
        self.assertRedirects(self.client.get("/accounts/"), "/accounts/login/")

        self.client.login(email=user.email)

        self.assertContains(
            self.client.get("/"), 'href="/accounts/logout/"', 1, status_code=200
        )
        self.assertRedirects(self.client.get("/accounts/"), "/accounts/update/")
        self.assertRedirects(self.client.get("/accounts/login/"), "/")

    def test_login_hint_removal_on_logout(self):
        """The login_hint cookie is removed when logging out explicitly"""
        user = factories.UserFactory.create()
        self.client.login(email=user.email)
        self.client.cookies.load({"login_hint": "test@example.com"})

        self.assertEqual(
            self.client.cookies.get("login_hint").value, "test@example.com"
        )
        response = self.client.get("/accounts/logout/")
        self.assertRedirects(response, "/accounts/login/")

        self.assertEqual(self.client.cookies.get("login_hint").value, "")

    def test_server_flow(self):
        """Exercise the OAuth2 webserver flow implementation"""
        FakeFlow.EMAIL = "user@example.com"
        FakeFlow.RAISE_EXCEPTION = False

        client = Client()
        response = client.get("/accounts/oauth2/")
        self.assertRedirects(
            response, "http://example.com/auth/", fetch_redirect_response=False
        )

        response = client.get("/accounts/oauth2/?code=x", HTTP_ACCEPT_LANGUAGE="en")
        self.assertRedirects(response, "/accounts/login/?error=1")
        self.assertEqual(
            messages(response),
            ["No user with email address user@example.com found."],
        )

        FakeFlow.EMAIL = "user@{}".format(settings.WORKBENCH.SSO_DOMAIN)
        response = client.get("/accounts/oauth2/?code=x", HTTP_ACCEPT_LANGUAGE="en")
        self.assertRedirects(response, "/accounts/update/")
        self.assertEqual(messages(response), ["Welcome! Please fill in your details."])
        response = client.post(
            "/accounts/update/",
            {
                "_full_name": "Test",
                "_short_name": "T",
                "language": "en",
                "working_time_model": factories.WorkingTimeModelFactory.create().pk,
            },
        )
        self.assertEqual(client.cookies.get("login_hint").value, FakeFlow.EMAIL)

        client = Client()
        response = client.get("/accounts/oauth2/?code=x")
        self.assertRedirects(response, "/")
        self.assertEqual(messages(response), [])
        self.assertEqual(client.cookies.get("login_hint").value, FakeFlow.EMAIL)

        # Disabled user
        User.objects.update(is_active=False)
        client = Client()
        FakeFlow.EMAIL = "user@{}".format(settings.WORKBENCH.SSO_DOMAIN)
        response = client.get("/accounts/oauth2/?code=x", HTTP_ACCEPT_LANGUAGE="en")
        self.assertRedirects(response, "/accounts/login/?error=1")
        self.assertEqual(
            messages(response),
            ["The user with email address user@feinheit.ch is inactive."],
        )

        client = Client()
        client.cookies.load({"login_hint": FakeFlow.EMAIL})

        FakeFlow.EMAIL = "user@example.com"
        response = client.get("/accounts/oauth2/?code=x", HTTP_ACCEPT_LANGUAGE="en")
        self.assertRedirects(response, "/accounts/login/?error=1")
        self.assertEqual(
            messages(response), ["No user with email address user@example.com found."]
        )
        # Login hint cookie has been removed when login fails
        self.assertEqual(client.cookies.get("login_hint").value, "")

    def test_server_flow_user_data_failure(self):
        """Failing to fetch user data shouldn't produce an internal server error"""
        FakeFlow.EMAIL = "user@example.com"
        FakeFlow.RAISE_EXCEPTION = True

        response = self.client.get("/accounts/oauth2/?code=x")
        self.assertRedirects(response, "/accounts/login/")
        self.assertEqual(
            messages(response),
            ["Fehler während Abholen der Daten. Bitte nochmals versuchen."],
        )

    def test_accounts_update_404(self):
        """No authentication and no saved email address --> 404"""
        response = self.client.get("/accounts/update/")
        self.assertEqual(response.status_code, 404)

    def test_account_update(self):
        """Users can only update some fields themselves"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        wtm = factories.WorkingTimeModelFactory.create()

        response = self.client.post(
            "/accounts/update/",
            {
                "_full_name": "Test",
                "_short_name": "T",
                "language": "en",
                "working_time_model": wtm.pk,
                "planning_hours_per_day": 6,
            },
        )
        self.assertRedirects(response, "/")

        user.refresh_from_db()
        self.assertEqual(user._short_name, "T")
        self.assertNotEqual(user.working_time_model, wtm)  # Can only be set initially



from workbench.timer import models
from workbench.tools import admin


@admin.register(models.Timestamp)
class TimestampAdmin(admin.ModelAdmin):
    date_hierarchy = "created_at"
    list_display = ["user", "created_at", "type", "notes"]
    list_filter = [("user", admin.RelatedOnlyFieldListFilter)]
    radio_fields = {"type": admin.HORIZONTAL}
    raw_id_fields = ["logged_hours", "logged_break"]



# Generated by Django 3.0.5 on 2020-05-04 11:38

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("timer", "0005_timestamp_project"),
    ]

    operations = [
        migrations.DeleteModel(
            name="TimerState",
        ),
    ]



# Generated by Django 2.2.2 on 2019-06-30 20:19

from django.db import migrations


def forwards(apps, schema_editor):
    Model = apps.get_model("projects.Project")
    for instance in Model.objects.select_related("customer", "contact"):
        instance._fts = " ".join(
            str(part)
            for part in [
                "%s-%04d" % (instance.created_at.year, instance._code),
                instance.customer.name,
                instance.contact.given_name if instance.contact else "",
                instance.contact.family_name if instance.contact else "",
            ]
        )
        instance.save()


class Migration(migrations.Migration):

    dependencies = [("projects", "0012_project__fts")]

    operations = [migrations.RunPython(forwards)]



from decimal import Decimal

from django import forms
from django.db.models import ProtectedError
from django.test import TestCase

from workbench import factories
from workbench.awt.models import Year  # any tools.Model()
from workbench.contacts.models import Organization
from workbench.projects.models import Project
from workbench.tools import formats
from workbench.tools.forms import Autocomplete
from workbench.tools.models import ModelWithTotal
from workbench.tools.testing import messages
from workbench.tools.validation import is_title_specific


class ToolsTest(TestCase):
    def test_model(self):
        """Shared methods of tools.models.Model work"""
        self.assertEqual(Year().code, "")
        self.assertEqual(Year(pk=3).code, "00003")
        self.assertEqual(Year().pretty_status, "")

    def test_invalid_autocomplete(self):
        """Invalid initial values for the Autocomplete widget are ignored"""

        class ProjectForm(forms.ModelForm):
            class Meta:
                model = Project
                fields = ["customer"]
                widgets = {"customer": Autocomplete(model=Organization)}

        form = ProjectForm(initial={"customer": "test"})
        self.assertIn('value=""', str(form))

    def test_model_with_total(self):
        """The calculation of totals excl. and incl. tax work"""
        m = ModelWithTotal(
            subtotal=Decimal("20"), discount=Decimal("5"), tax_rate=Decimal("8.0")
        )
        m._calculate_total()

        self.assertAlmostEqual(m._round_5cents(Decimal("1.02")), Decimal("1.00"))
        self.assertAlmostEqual(m._round_5cents(Decimal("1.03")), Decimal("1.05"))
        self.assertAlmostEqual(m.tax_amount, Decimal("1.20"))
        self.assertAlmostEqual(m.total_excl_tax, Decimal("15"))
        self.assertAlmostEqual(m.total, Decimal("16.20"))

        self.assertEqual(m.pretty_total_excl, "15.00 excl. tax (5.00 discount)")

    def test_create_absence_redirect(self):
        """The create view uses the get_redirect_url of models"""
        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/absences/create/")
        self.assertRedirects(response, "/absences/")

    def test_deletion(self):
        """Related objects are checked automatically when attempting deletion"""
        hours = factories.LoggedHoursFactory.create(description="Bla")
        project = hours.service.project

        with self.assertRaises(ProtectedError):
            project.delete()

        self.client.force_login(hours.rendered_by)
        response = self.client.get(project.urls["delete"])
        self.assertRedirects(response, project.urls["detail"])
        self.assertEqual(
            messages(response),
            [
                "Cannot delete '{}' because of related objects (Any service: Bla)."
                "".format(project)
            ],
        )

    def test_formats_hours(self):
        """Number formatting"""
        for value, result in [
            (Decimal("42.22"), "42.2h"),
            (Decimal("42.27"), "42.3h"),
            (Decimal("3"), "3.0h"),
            (Decimal("-123"), "-123.0h"),
            (Decimal("0.003"), "0.0h"),
            (Decimal("-0.003"), "0.0h"),
            (Decimal("12345"), "12’345.0h"),
            (None, "0.0h"),
        ]:
            with self.subTest(value=value, result=result):
                self.assertEqual(formats.hours(value), result)

    def test_is_title_specific(self):
        """is_title_specific tests"""
        self.assertTrue(is_title_specific("Implementation Kontaktformular"))
        self.assertFalse(is_title_specific("Programmierung allgemein"))
        self.assertTrue(
            is_title_specific("Analyse der Storyboards und allgemeine Anforderungen")
        )

    def test_formats_minutes_and_hours(self):
        """Duration formatting"""
        for value, result in [
            (0, "0 minutes"),
            (60, "1 minute"),
            (3600, "1 hour 0 minutes"),
            (7261, "2 hours 1 minute"),
        ]:
            with self.subTest(value=value, result=result):
                self.assertEqual(formats.hours_and_minutes(value), result)



# Generated by Django 2.1.7 on 2019-03-09 09:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("projects", "0003_auto_20190306_2029")]

    operations = [
        migrations.RemoveField(model_name="project", name="invoicing"),
        migrations.RemoveField(model_name="project", name="maintenance"),
        migrations.RemoveField(model_name="project", name="status"),
        migrations.AddField(
            model_name="project",
            name="closed_on",
            field=models.DateField(blank=True, null=True, verbose_name="closed on"),
        ),
        migrations.AddField(
            model_name="project",
            name="type",
            field=models.CharField(
                choices=[
                    ("acquisition", "Acquisition"),
                    ("maintenance", "Maintenance"),
                    ("order", "Order"),
                    ("internal", "Internal"),
                ],
                default="order",
                max_length=20,
                verbose_name="type",
            ),
            preserve_default=False,
        ),
        migrations.AlterModelOptions(
            name="service",
            options={
                "ordering": ["position", "created_at"],
                "verbose_name": "service",
                "verbose_name_plural": "services",
            },
        ),
    ]



# Generated by Django 3.0.2 on 2020-01-13 14:54

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("accounts", "0006_auto_20191207_1307"),
    ]

    operations = [
        migrations.CreateModel(
            name="Team",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100, verbose_name="name")),
                (
                    "members",
                    models.ManyToManyField(
                        related_name="teams",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="members",
                    ),
                ),
            ],
            options={
                "verbose_name": "team",
                "verbose_name_plural": "teams",
                "ordering": ["name"],
            },
        ),
    ]



import datetime as dt
import math
from itertools import groupby

from django import template
from django.db import models
from django.db.models import Sum
from django.template.defaultfilters import linebreaksbr
from django.urls import reverse
from django.utils.html import format_html, format_html_join, mark_safe
from django.utils.text import capfirst
from django.utils.translation import gettext as _

from workbench.logbook.models import LoggedHours
from workbench.notes.forms import NoteForm
from workbench.notes.models import Note
from workbench.projects.models import Service
from workbench.tools.formats import Z1, Z2, currency, days, hours, local_date_format
from workbench.tools.forms import querystring as _qs
from workbench.tools.history import EVERYTHING


register = template.Library()


register.filter(currency)
register.filter(days)
register.filter(hours)


@register.filter(name="local_date_format")
def local_date_format_filter(dttm, fmt=None):
    return local_date_format(dttm, fmt=fmt)


@register.simple_tag
def link_or_none(object, pretty=None, none=mark_safe("&ndash;"), with_badge=False):
    if object == 0:
        return object
    elif not object:
        return none
    elif hasattr(object, "html_link"):
        return object.html_link()
    elif hasattr(object, "get_absolute_url"):
        return format_html(
            '<a href="{}">{}{}</a>',
            object.get_absolute_url(),
            h(pretty or object),
            format_html(" {}", object.status_badge)
            if with_badge and hasattr(object, "status_badge")
            else "",
        )
    return pretty or object


@register.filter
def field_value_pairs(object, fields=None):
    fields = (
        fields.split(",")
        if fields
        else getattr(object, "field_value_pairs", EVERYTHING)
    )
    for field in object._meta.get_fields():
        if (
            field.one_to_many
            or field.one_to_one
            or field.many_to_many
            or field.primary_key
            or field.name in {"_fts"}
            or field.name not in fields
        ):
            continue

        if field.choices:
            yield (capfirst(field.verbose_name), object._get_FIELD_display(field))

        elif isinstance(field, models.TextField):
            yield (
                capfirst(field.verbose_name),
                linebreaksbr(getattr(object, field.name)),
            )

        else:
            value = getattr(object, field.name)
            if isinstance(value, dt.date):
                value = local_date_format(value)
            elif isinstance(value, bool):
                value = _("yes") if value else _("no")

            yield (capfirst(field.verbose_name), value)


@register.filter
def h(object):
    if hasattr(object, "__html__"):
        return object.__html__()
    return object


@register.filter
def group_hours_by_day(iterable):
    for day, instances in groupby(iterable, lambda logged: logged.rendered_on):
        instances = list(instances)
        yield (day, sum((item.hours for item in instances), Z1), instances)


def deal_group(deal):
    decision_expected_within = (
        (deal.decision_expected_on - dt.date.today()).days
        if deal.decision_expected_on
        else 99999999999
    )

    if deal.probability == deal.HIGH:
        if decision_expected_within <= 4 * 7:
            return (1, _("high probability, next 4 weeks"))
        elif decision_expected_within <= 12 * 7:
            return (2, _("high probability, next 12 weeks"))
        else:
            return (3, _("high probability, later"))
    elif deal.probability == deal.NORMAL:
        return (4, _("normal probability"))
    else:
        return (5, _("unknown probability"))


@register.filter
def group_deals_by_probability(iterable, should_group):
    if not should_group:
        deals = list(iterable)
        yield {
            "title": capfirst(iterable.model._meta.verbose_name_plural),
            "deals": deals,
            "sum": sum((deal.value for deal in deals), Z2),
        }
        return

    for group, deals in groupby(iterable, deal_group):
        deals = list(deals)
        yield {
            "title": group[1],
            "deals": deals,
            "sum": sum((deal.value for deal in deals), Z2),
        }


@register.simple_tag
def bar(value, one):
    if not one:
        return ""

    percentage = int(100 * value / one)

    bars = [("bg-success", min(75, max(0, percentage)))]

    if percentage >= 75:
        bars.append(("bg-caveat", min(25, percentage - 75)))
    if percentage > 100:
        bars.append(("bg-danger", percentage - 100))
        bars = [(cls, round(part * 100 / percentage, 2)) for cls, part in bars]

    return format_html(
        '<div class="progress progress-line" title="{percentage}%">{bars}</div>',
        percentage=percentage,
        bars=format_html_join(
            "",
            '<div class="progress-bar {}" role="progressbar" style="width:{}%"></div>',
            bars,
        ),
    )


@register.simple_tag
def pie(value, one, size=20, type="bad"):
    if not one:
        angle = 0
    else:
        angle = 2 * math.pi * min(0.999, float(value / one))

    hsize = size // 2

    return format_html(
        """\
<svg width="{size}" height="{size}" class="pie {type}" style="display: inline-block">
  <circle r="{hsize}" cx="{hsize}" cy="{hsize}" class="pie-circle" />
  <path d="M {hsize} 0 A {hsize} {hsize} 0 {large_arc} 1 {x} {y} L {hsize} {hsize} z" class="pie-arc" />
</svg>""",  # noqa
        large_arc=1 if angle > math.pi else 0,
        x=hsize + math.sin(angle) * hsize,
        y=hsize - math.cos(angle) * hsize,
        size=size + 2,
        hsize=hsize,
        type=type,
    )


@register.simple_tag(takes_context=True)
def querystring(context, **kwargs):
    return _qs(context["request"].GET, **kwargs)


@register.simple_tag(takes_context=True)
def page_links(context, page_obj):
    return (
        (page, querystring(context, page=page))
        for page in page_obj.paginator.page_range
        if abs(page - page_obj.number) < 7
    )


@register.simple_tag
def history_link(instance):
    return (
        format_html(
            """\
<a href="{}"
    class="tiny-icons"
    data-toggle="ajaxmodal"
    title="{}">
<svg xmlns="http://www.w3.org/2000/svg" width="14" height="16" viewBox="0 0 14 16"><path fill-rule="evenodd" d="M13 3H7c-.55 0-1 .45-1 1v8c0 .55.45 1 1 1h6c.55 0 1-.45 1-1V4c0-.55-.45-1-1-1zm-1 8H8V5h4v6zM4 4h1v1H4v6h1v1H4c-.55 0-1-.45-1-1V5c0-.55.45-1 1-1zM1 5h1v1H1v4h1v1H1c-.55 0-1-.45-1-1V6c0-.55.45-1 1-1z"/></svg>
</a>""",  # noqa
            reverse("history", args=(instance._meta.db_table, "id", instance.pk)),
            _("History"),
        )
        if instance.pk
        else ""
    )


@register.simple_tag
def project_statistics_row(project_logged_hours, service_logged_hours):
    service = dict(service_logged_hours)
    return [(user, service.get(user)) for user, _ in project_logged_hours]


@register.simple_tag
def percentage(value, one):
    return (
        format_html(
            '<span class="font-weight-normal text-black-30">{}%</span>',
            round(100 * value / one),
        )
        if one
        else ""
    )


@register.filter
def has_feature(user, feature):
    if not hasattr(user, "features"):  # e.g. AnonymousUser
        return False
    return user.features[feature]


@register.filter
def label(instance, field):
    return capfirst(instance._meta.get_field(field).verbose_name)


@register.filter
def addf(a, b):
    """Add values without converting them to integers (as |add seems to do)"""
    return a + b


@register.inclusion_tag("notes/widget.html", takes_context=True)
def notes(context, instance):
    request = context["request"]
    notes = (
        Note.objects.for_content_object(instance).select_related("created_by").reverse()
    )
    return {
        "form": NoteForm(request=request, content_object=instance),
        "notes": notes,
        "request": request,
    }


@register.filter
def analyze_projects(object_list):
    pks = [object.pk for object in object_list]
    service_hours = {
        row["project"]: row["service_hours__sum"]
        for row in Service.objects.budgeted()
        .filter(project__in=pks)
        .order_by()
        .values("project")
        .annotate(Sum("service_hours"))
    }
    logged_hours = {
        row["service__project"]: row["hours__sum"]
        for row in LoggedHours.objects.filter(service__project__in=pks)
        .order_by()
        .values("service__project")
        .annotate(Sum("hours"))
    }

    for object in object_list:
        object.analyzed = {
            "service_hours": service_hours.get(object.pk, 0),
            "logged_hours": logged_hours.get(object.pk, 0),
        }
    return object_list



# Generated by Django 3.0 on 2019-12-10 14:37

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("logbook", "0015_remove_loggedcost_project"),
    ]

    operations = [
        migrations.AddField(
            model_name="loggedcost",
            name="expense_cost",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="original cost",
            ),
        ),
        migrations.AddField(
            model_name="loggedcost",
            name="expense_currency",
            field=models.CharField(
                blank=True, max_length=3, verbose_name="original currency"
            ),
        ),
    ]



from django.contrib import messages
from django.http import HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.translation import gettext as _

from workbench import generic
from workbench.projects.forms import OffersRenumberForm, ProjectAutocompleteForm
from workbench.projects.models import Project, Service
from workbench.services.models import ServiceType
from workbench.templatetags.workbench import h


def select(request):
    if not request.is_ajax():
        return redirect("/")
    form = ProjectAutocompleteForm(request.POST if request.method == "POST" else None)
    if form.is_valid():
        data = form.cleaned_data
        return JsonResponse(
            {
                "redirect": data["service"].get_absolute_url()
                if data["service"]
                else data["project"].get_absolute_url()
            },
            status=299,
        )
    return render(
        request,
        "generic/select_object.html",
        {
            "form": form,
            "title": _("Jump to project"),
            "links": [
                {"title": h(project), "url": project.get_absolute_url(), "attrs": ""}
                for project in request.user.active_projects
            ],
        },
    )


class OffersRenumberView(generic.UpdateView):
    form_class = OffersRenumberForm


def assign_service_type(request, pk):
    service = Service.objects.get(pk=pk)
    service_type = ServiceType.objects.get(pk=request.GET.get("service_type"))

    service.effort_type = service_type.title
    service.effort_rate = service_type.hourly_rate
    service.save()
    messages.success(
        request,
        _("%(class)s '%(object)s' has been updated successfully.")
        % {"class": service._meta.verbose_name, "object": service},
    )
    return redirect(service)


def set_order(request):
    for index, id in enumerate(request.POST.getlist("ids[]")):
        Service.objects.filter(id=id).update(position=10 * (index + 1))
    return HttpResponse("OK", status=202)  # Accepted


def services(request, pk):
    project = get_object_or_404(Project, pk=pk)
    offers = {None: {"label": _("Not offered yet"), "options": []}}
    for service in project.services.logging().select_related("offer__owned_by"):
        if service.offer not in offers:
            offers[service.offer] = {"label": str(service.offer), "options": []}
        offers[service.offer]["options"].append(
            {"label": str(service), "value": service.id}
        )

    return JsonResponse(
        {
            "id": project.id,
            "code": project.code,
            "title": project.title,
            "owned_by": project.owned_by.get_short_name(),
            "customer_id": project.customer_id,
            "services": [
                data for offer, data in sorted(offers.items()) if data["options"]
            ],
        }
    )


def projects(request):
    return JsonResponse(
        {
            "projects": [
                {"label": str(project), "value": project.pk}
                for project in request.user.active_projects
            ],
        }
    )



# Generated by Django 2.1.7 on 2019-03-11 13:56

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("invoices", "0006_auto_20190309_2143")]

    operations = [
        migrations.RemoveField(model_name="cost", name="service"),
        migrations.RemoveField(model_name="effort", name="service"),
        migrations.RemoveField(model_name="effort", name="service_type"),
        migrations.AddField(
            model_name="service",
            name="effort_rate",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="hourly rate",
            ),
        ),
        migrations.AddField(
            model_name="service",
            name="effort_type",
            field=models.CharField(
                blank=True, max_length=50, verbose_name="effort type"
            ),
        ),
        migrations.AddField(
            model_name="service",
            name="service_cost",
            field=models.DecimalField(
                decimal_places=2,
                default=0,
                max_digits=10,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="service cost",
            ),
        ),
        migrations.AddField(
            model_name="service",
            name="service_hours",
            field=models.DecimalField(
                decimal_places=1,
                default=0,
                max_digits=4,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="service hours",
            ),
        ),
        migrations.AddField(
            model_name="service",
            name="third_party_costs",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                default=None,
                help_text="Total incl. tax for third-party services.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="third party costs",
            ),
        ),
        migrations.AlterField(
            model_name="service",
            name="cost",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="cost",
            ),
        ),
        migrations.AlterField(
            model_name="service",
            name="effort_hours",
            field=models.DecimalField(
                blank=True,
                decimal_places=1,
                max_digits=4,
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="hours",
            ),
        ),
        migrations.DeleteModel(name="Cost"),
        migrations.DeleteModel(name="Effort"),
    ]



from django.urls import path, re_path

from workbench import generic
from workbench.accounts.forms import TeamForm, TeamSearchForm, UserSearchForm
from workbench.accounts.models import Team, User
from workbench.planning.views import TeamPlanningView, UserPlanningView

from . import views


urlpatterns = [
    re_path(r"^accounts/$", views.accounts, name="accounts"),
    re_path(
        r"^accounts/update/$", views.UserUpdateView.as_view(), name="accounts_update"
    ),
    re_path(r"^accounts/login/$", views.login, name="login"),
    re_path(r"^accounts/oauth2/$", views.oauth2, name="accounts_oauth2"),
    re_path(r"^accounts/logout/$", views.logout, name="logout"),
    #
    path(
        "users/",
        generic.ListView.as_view(
            model=User,
            search_form_class=UserSearchForm,
            show_create_button=False,
        ),
        name="accounts_user_list",
    ),
    path(
        "users/<int:pk>/",
        generic.DetailView.as_view(model=User),
        name="accounts_user_detail",
    ),
    path(
        "users/<int:pk>/planning/",
        UserPlanningView.as_view(),
        name="accounts_user_planning",
    ),
    path(
        "users/<int:pk>/statistics/",
        views.ProfileView.as_view(),
        name="accounts_user_statistics",
    ),
    # Teams
    re_path(
        r"^teams/$",
        generic.ListView.as_view(model=Team, search_form_class=TeamSearchForm),
        name="accounts_team_list",
    ),
    re_path(
        r"^teams/(?P<pk>\d+)/$",
        generic.DetailView.as_view(model=Team),
        name="accounts_team_detail",
    ),
    re_path(
        r"^teams/create/$",
        generic.CreateView.as_view(form_class=TeamForm, model=Team),
        name="accounts_team_create",
    ),
    re_path(
        r"^teams/(?P<pk>\d+)/update/$",
        generic.UpdateView.as_view(form_class=TeamForm, model=Team),
        name="accounts_team_update",
    ),
    path(
        "teams/<int:pk>/planning/",
        TeamPlanningView.as_view(),
        name="accounts_team_planning",
    ),
    re_path(
        r"^teams/(?P<pk>\d+)/delete/$",
        generic.DeleteView.as_view(model=Team),
        name="accounts_team_delete",
    ),
]



# Generated by Django 2.1.7 on 2019-03-21 09:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contacts", "0004_person_is_archived")]

    operations = [
        migrations.AddField(
            model_name="person",
            name="address_on_first_name_terms",
            field=models.BooleanField(
                default=False, verbose_name="address on first-name terms"
            ),
        ),
        migrations.AlterField(
            model_name="person",
            name="address",
            field=models.CharField(
                blank=True,
                help_text="Mr./Ms./...",
                max_length=100,
                verbose_name="address",
            ),
        ),
    ]



import datetime as dt
from decimal import Decimal

from django.core.exceptions import ValidationError
from django.test import TestCase
from django.test.utils import override_settings
from django.utils.translation import deactivate_all

from workbench import factories
from workbench.accounts.features import FEATURES
from workbench.awt.models import Absence, Employment
from workbench.awt.reporting import active_users, annual_working_time
from workbench.awt.utils import monthly_days
from workbench.tools.forms import WarningsForm
from workbench.tools.testing import check_code, messages
from workbench.tools.validation import in_days


class AWTTest(TestCase):
    def setUp(self):
        deactivate_all()

    def test_redirect(self):
        """Invalid year produces a redirect"""
        user = factories.UserFactory.create()
        self.client.force_login(user)
        response = self.client.get("/report/annual-working-time/?year=asdf")
        self.assertRedirects(response, "/report/annual-working-time/")

    def test_year(self):
        """Basic year methods work"""
        year = factories.YearFactory.create()
        self.assertEqual(year.months, [30 for i in range(12)])

    def test_monthly_days(self):
        """Days per month are calculated correctly  for partial months"""
        for from_, until_, dates in [
            (
                dt.date(2018, 2, 1),
                dt.date(2018, 7, 31),
                [
                    (dt.date(2018, 2, 1), 28),
                    (dt.date(2018, 3, 1), 31),
                    (dt.date(2018, 4, 1), 30),
                    (dt.date(2018, 5, 1), 31),
                    (dt.date(2018, 6, 1), 30),
                    (dt.date(2018, 7, 1), 31),
                ],
            ),
            (
                dt.date(2018, 2, 2),
                dt.date(2018, 7, 30),
                [
                    (dt.date(2018, 2, 1), 27),
                    (dt.date(2018, 3, 1), 31),
                    (dt.date(2018, 4, 1), 30),
                    (dt.date(2018, 5, 1), 31),
                    (dt.date(2018, 6, 1), 30),
                    (dt.date(2018, 7, 1), 30),
                ],
            ),
            (dt.date(2018, 2, 10), dt.date(2018, 2, 19), [(dt.date(2018, 2, 1), 10)]),
        ]:
            with self.subTest(locals()):
                self.assertEqual(list(monthly_days(from_, until_)), dates)

    def test_working_time(self):
        """Verify that the annual working time report yields correct results"""
        service = factories.ServiceFactory.create()
        user = service.project.owned_by
        year = factories.YearFactory.create(
            year=2018, working_time_model=user.working_time_model
        )
        user.loggedhours.create(
            service=service,
            created_by=user,
            hours=1000,
            description="anything",
            rendered_on=dt.date(2018, 1, 1),
        )
        user.absences.create(starts_on=dt.date(2018, 1, 1), days=5, reason="vacation")
        user.absences.create(starts_on=dt.date(2018, 4, 1), days=45, reason="vacation")
        user.absences.create(starts_on=dt.date(2018, 7, 1), days=10, reason="paid")
        user.absences.create(starts_on=dt.date(2018, 10, 1), days=10, reason="vacation")
        user.employments.create(
            date_from=dt.date(2014, 1, 1),
            date_until=dt.date(2014, 3, 31),
            percentage=80,
            vacation_weeks=5,
        )
        user.employments.create(
            date_from=dt.date(2017, 1, 1), percentage=80, vacation_weeks=5
        )
        user.employments.create(
            date_from=dt.date(2018, 10, 1), percentage=100, vacation_weeks=5
        )

        # Should have no effect
        a = user.absences.create(starts_on=dt.date(2018, 8, 1), days=10, reason="other")
        self.assertFalse(a.is_working_time)

        employments = list(user.employments.all())
        self.assertEqual(employments[0].date_until, dt.date(2014, 3, 31))
        self.assertEqual(employments[1].date_until, dt.date(2018, 9, 30))
        self.assertEqual(employments[2].date_until, dt.date(9999, 12, 31))

        awt = annual_working_time(year.year, users=[user])["statistics"][0]

        self.assertAlmostEqual(awt["totals"]["target_days"], Decimal("360"))
        self.assertAlmostEqual(awt["totals"]["percentage"], Decimal("85"))
        # 3/4 * 20 + 1/4 * 25 = 21.25
        self.assertAlmostEqual(
            awt["totals"]["available_vacation_days"], Decimal("21.25")
        )

        self.assertAlmostEqual(awt["totals"]["hours"], Decimal("1000"))
        self.assertAlmostEqual(awt["totals"]["absence_vacation"], Decimal("60"))
        # 21.25 - 50 - 10 = -38.75
        self.assertAlmostEqual(
            awt["totals"]["vacation_days_correction"], Decimal("-38.75")
        )
        self.assertAlmostEqual(awt["totals"]["absence_sickness"], Decimal("0"))
        self.assertAlmostEqual(awt["totals"]["absence_paid"], Decimal("10"))

        # 3/4 * 80% * 360 + 1/4 * 100% * 360 = 306
        self.assertAlmostEqual(awt["totals"]["target"], Decimal("306") * 8)

        # 1000 / 8 + 21.25 + 10 = 125 + 21.25 + 10 = 156.25
        self.assertAlmostEqual(awt["totals"]["working_time"], Decimal("156.25") * 8)

        # 306 * 8 - 1000 - 21.25 * 8 - 10 * 8 = 1198
        self.assertAlmostEqual(awt["totals"]["running_sum"], Decimal("-1198"))

        # Now, test that the PDF does not crash
        self.client.force_login(user)

        response = self.client.get("/report/annual-working-time/?export=pdf&year=2018")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/pdf")

        user.employments.all().delete()
        response = self.client.get("/report/annual-working-time/?export=pdf&year=2018")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/pdf")

        user.absences.create(
            starts_on=dt.date(2018, 10, 1), days=10, reason="correction"
        )
        response = self.client.get("/report/annual-working-time/?export=pdf&year=2018")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/pdf")

    def test_admin_list(self):
        """The admin changelist of years contains the calculated sum of working days"""
        self.client.force_login(factories.UserFactory.create(is_admin=True))
        factories.YearFactory.create()
        response = self.client.get("/admin/awt/year/")
        self.assertContains(response, '<td class="field-days">360.00</td>')

    def test_absence_editing(self):
        """Creating and updating absences in the current year works"""
        user = factories.UserFactory.create()
        self.client.force_login(user)
        response = self.client.post(
            "/absences/create/",
            {
                "modal-user": user.pk,
                "modal-starts_on": dt.date.today().isoformat(),
                "modal-days": 3,
                "modal-description": "Sick",
                "modal-reason": "sickness",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)
        absence = Absence.objects.get()

        response = self.client.get(
            absence.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)

        Absence.objects.update(starts_on=dt.date(2018, 1, 1))
        response = self.client.get(
            absence.urls["update"], HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Absences of past years are locked.")

        response = self.client.post(
            "/absences/create/",
            {
                "modal-user": user.pk,
                "modal-starts_on": "2018-01-01",
                "modal-days": 3,
                "modal-description": "Sick",
                "modal-reason": "sickness",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(
            response, "Creating absences for past years is not allowed."
        )

    def test_no_data_no_crash(self):
        """POSTing the absence form without any data does not crash the backend"""
        user = factories.UserFactory.create()
        self.client.force_login(user)
        response = self.client.post(
            "/absences/create/", {}, HTTP_X_REQUESTED_WITH="XMLHttpRequest"
        )
        self.assertEqual(response.status_code, 200)

    def test_planning_skills(self):
        """Creating absences in the far future requires impressive-planning-skills"""
        user = factories.UserFactory.create()
        self.client.force_login(user)
        response = self.client.post(
            "/absences/create/",
            {
                "modal-user": user.pk,
                "modal-starts_on": "2100-01-01",
                "modal-days": 3,
                "modal-description": "Sick",
                "modal-reason": "sickness",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(
            response,
            "Impressive planning skills or wrong date?"
            " Absence starts in more than one year.",
        )

        response = self.client.post(
            "/absences/create/",
            {
                "modal-user": user.pk,
                "modal-starts_on": "2100-01-01",
                "modal-days": 3,
                "modal-description": "Sick",
                "modal-reason": "sickness",
                WarningsForm.ignore_warnings_id: "impressive-planning-skills",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

    def test_employment_model(self):
        """Exercise the employment model"""
        factories.UserFactory.create()  # Additional instance for active_users()

        year = factories.YearFactory.create()
        user = factories.UserFactory.create()

        employment = Employment.objects.create(
            user=user, percentage=100, vacation_weeks=5, date_from=dt.date(2018, 1, 1)
        )
        self.assertEqual(employment.date_until, dt.date.max)
        self.assertEqual(str(employment), "since 01.01.2018")

        self.assertEqual(list(active_users(year.year)), [user])

        employment.date_until = dt.date(2018, 6, 30)
        employment.save()
        self.assertEqual(str(employment), "01.01.2018 - 30.06.2018")

        self.assertEqual(list(active_users(year.year)), [])

    def test_report_view(self):
        """The export view produces individual PDFs or ZIP files containing PDFs"""
        year = factories.YearFactory.create()
        user = factories.UserFactory.create(
            _full_name="Fritz", working_time_model=year.working_time_model
        )
        inactive = factories.UserFactory.create(
            working_time_model=year.working_time_model
        )
        Employment.objects.create(user=user, percentage=50, vacation_weeks=5)

        # New user has a different working time model...
        self.client.force_login(factories.UserFactory.create(_full_name="Hans"))

        url = "/report/annual-working-time/"
        response = self.client.get(url)
        self.assertNotContains(response, str(user))
        self.assertNotContains(response, str(inactive))

        response = self.client.get(url + "?user=active")
        self.assertContains(response, str(user))
        self.assertNotContains(response, str(inactive))

        response = self.client.get(url + "?user=" + str(inactive.pk))
        self.assertContains(response, str(inactive))

        response = self.client.get(url + "?year=2018")
        self.assertEqual(
            messages(response),
            [
                "No annual working time defined for user Hans"
                " with working time model Test."
            ],
        )

        other = factories.UserFactory.create(working_time_model=year.working_time_model)
        Employment.objects.create(user=other, percentage=50, vacation_weeks=5)
        factories.AbsenceFactory.create(user=other)

        response = self.client.get(url + "?export=pdf&user=active")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/zip")

        response = self.client.get(url + "?export=pdf&user={}".format(user.pk))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response["content-type"], "application/pdf")

    def test_non_ajax_redirect(self):
        """The absence URL redirects when encountering a non-AJAX request"""
        absence = factories.AbsenceFactory.create()
        self.client.force_login(absence.user)
        response = self.client.get(absence.urls["detail"])
        self.assertRedirects(response, "/absences/?u=" + str(absence.user.pk))

    def test_list(self):
        """The absence list and its filters do not crash"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        code = check_code(self, "/absences/")
        code("")
        code("u=-1")
        code("u={}".format(user.pk))
        code("reason=sickness")
        code("reason=nothing", 302)
        code("export=xlsx")

    def test_employment_validation(self):
        """Test model validation of employments"""
        user = factories.UserFactory.create()
        kw = {"user": user, "percentage": 100, "vacation_weeks": 5, "notes": ""}
        Employment(**kw).full_clean()
        Employment(
            **kw, green_hours_target=20, hourly_labor_costs=Decimal(20)
        ).full_clean()

        msg = [
            (
                "__all__",
                [
                    "Either provide both hourly labor costs"
                    " and green hours target or none."
                ],
            )
        ]

        with self.assertRaises(ValidationError) as cm:
            Employment(**kw, green_hours_target=20).full_clean()
        self.assertEqual(list(cm.exception), msg)

        with self.assertRaises(ValidationError) as cm:
            Employment(**kw, hourly_labor_costs=20).full_clean()
        self.assertEqual(list(cm.exception), msg)

        msg = [("date_until", ["Employments cannot end before they began."])]

        with self.assertRaises(ValidationError) as cm:
            Employment(
                **kw, date_from=dt.date.today(), date_until=dt.date(2020, 1, 30)
            ).full_clean()
        self.assertEqual(list(cm.exception), msg)

    def test_calendar(self):
        """The absence calendar report does not crash"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        Absence.objects.create(
            user=user,
            starts_on=dt.date.today(),
            days=0,
            description="Test",
            reason=Absence.VACATION,
        )
        Absence.objects.create(
            user=user,
            starts_on=in_days(10),
            ends_on=in_days(20),
            days=0,
            description="Test",
            reason=Absence.VACATION,
        )

        code = check_code(self, "/report/absence-calendar/")
        code("")

        team = factories.TeamFactory.create()
        user.teams.add(team)
        code("team={}".format(team.pk))
        code("team=abc", status_code=302)

    def test_absence_validation(self):
        """Test model validation of absences"""
        user = factories.UserFactory.create()
        kw = {
            "user": user,
            "starts_on": dt.date(2020, 2, 1),
            "days": 0,
            "description": "Nothing",
            "reason": "vacation",
        }
        Absence(**kw).full_clean()
        Absence(**kw, ends_on=dt.date(2020, 2, 2)).full_clean()

        msg = [("ends_on", ["Absences cannot end before they began."])]

        with self.assertRaises(ValidationError) as cm:
            Absence(**kw, ends_on=dt.date(2020, 1, 30)).full_clean()
        self.assertEqual(list(cm.exception), msg)

        msg = [("ends_on", ["Start and end must be in the same year."])]

        with self.assertRaises(ValidationError) as cm:
            Absence(**kw, ends_on=dt.date(2022, 1, 30)).full_clean()
        self.assertEqual(list(cm.exception), msg)

    @override_settings(FEATURES={FEATURES.WORKING_TIME_CORRECTION: False})
    def test_correction(self):
        """Absences with reason "correction" only with bookkeeping"""
        user = factories.UserFactory.create()
        self.client.force_login(user)

        response = self.client.get(
            Absence.urls["create"],
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertNotContains(response, 'value="correction"')

        absence = factories.AbsenceFactory.create(user=user, reason=Absence.CORRECTION)
        response = self.client.get(
            absence.urls["update"],
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "You are not permitted to edit absences of type")

    def test_absences_without_enforce_same_week_logging(self):
        """Absences in past years without enforce_same_week_logging"""
        user = factories.UserFactory.create(enforce_same_week_logging=False)
        self.client.force_login(user)
        response = self.client.post(
            "/absences/create/",
            {
                "modal-user": user.pk,
                "modal-starts_on": dt.date(2020, 1, 1).isoformat(),
                "modal-days": 3,
                "modal-description": "Sick",
                "modal-reason": "sickness",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "past-years-absences")

        response = self.client.post(
            "/absences/create/",
            {
                "modal-user": user.pk,
                "modal-starts_on": dt.date(2020, 1, 1).isoformat(),
                "modal-days": 3,
                "modal-description": "Sick",
                "modal-reason": "sickness",
                WarningsForm.ignore_warnings_id: "past-years-absences",
            },
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 201)

        absence = Absence.objects.get()
        response = self.client.get(
            absence.urls["update"],
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertContains(response, "<input")



def substitute_with(*, to_delete, keep):
    """
    Substitute the first argument with the second in all relations,
    and delete the first argument afterwards.
    """
    assert to_delete.__class__ == keep.__class__
    assert to_delete.pk != keep.pk

    fields = [
        f
        for f in to_delete._meta.get_fields()
        if (f.one_to_many or f.one_to_one) and f.auto_created and not f.concrete
    ]

    for related_object in fields:
        queryset = related_object.related_model._base_manager.complex_filter(
            {related_object.field.name: to_delete.pk}
        )

        queryset.update(**{related_object.field.name: keep.pk})
    to_delete.delete()



import itertools
import re
from urllib.parse import urlencode

from django.db import models
from django.db.models import Sum
from django.utils.functional import cached_property
from django.utils.translation import gettext, gettext_lazy as _

import phonenumbers
from django_countries.fields import CountryField

from workbench.accounts.models import User
from workbench.tools.formats import Z2
from workbench.tools.models import Model, SearchQuerySet
from workbench.tools.urls import model_urls
from workbench.tools.validation import raise_if_errors


@model_urls
class Group(Model):
    title = models.CharField(_("title"), max_length=100)

    class Meta:
        ordering = ("title",)
        verbose_name = _("group")
        verbose_name_plural = _("groups")

    def __str__(self):
        return self.title


class OrganizationQuerySet(SearchQuerySet):
    def active(self):
        return self.filter(is_archived=False)


@model_urls
class Organization(Model):
    is_archived = models.BooleanField(
        _("is archived"),
        default=False,
        help_text=_("It is preferrable to archive records instead of deleting them."),
    )
    name = models.TextField(_("name"))
    is_private_person = models.BooleanField(_("is private person"), default=False)
    notes = models.TextField(_("notes"), blank=True)
    primary_contact = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("primary contact"),
    )
    default_billing_address = models.TextField(
        _("default billing address"),
        blank=True,
        help_text=_(
            "Mainly useful for organizations with a centralized"
            " accounts payable departement."
        ),
    )
    groups = models.ManyToManyField(Group, verbose_name=_("groups"), blank=True)

    objects = OrganizationQuerySet.as_manager()

    class Meta:
        ordering = ("name",)
        verbose_name = _("organization")
        verbose_name_plural = _("organizations")

    def __str__(self):
        parts = [self.name.replace("\n", " ")]
        if self.is_archived:
            parts.append(gettext("(archived)"))
        return " ".join(parts)

    @classmethod
    def allow_delete(cls, instance, request):
        return None

    def recent_projects(self):
        return self.project_set.select_related("owned_by")[:10]

    def recent_invoices(self):
        return self.invoice_set.select_related("owned_by")[:10]

    def recent_offers(self):
        from workbench.offers.models import Offer

        return Offer.objects.filter(project__customer=self).select_related(
            "project", "owned_by"
        )[:10]

    @cached_property
    def statistics(self):
        from workbench.reporting.project_budget_statistics import (
            project_budget_statistics,
        )

        pbs = project_budget_statistics(self.project_set.all())["overall"]

        invoices = (
            self.invoice_set.invoiced()
            .filter(project__isnull=True)
            .order_by()
            .aggregate(Sum("total_excl_tax"), Sum("third_party_costs"))
        )

        pbs["invoiced"] += invoices["total_excl_tax__sum"] or Z2
        pbs["third_party_costs"] += invoices["third_party_costs__sum"] or Z2

        pbs["gross_margin_per_hour"] = (
            (pbs["invoiced"] - pbs["third_party_costs"]) / pbs["hours"]
            if pbs["hours"]
            else None
        )
        return pbs


class PersonQuerySet(SearchQuerySet):
    def active(self):
        return self.filter(is_archived=False)


@model_urls
class Person(Model):
    is_archived = models.BooleanField(
        _("is archived"),
        default=False,
        help_text=_("It is preferrable to archive records instead of deleting them."),
    )
    given_name = models.CharField(_("given name"), max_length=100)
    family_name = models.CharField(_("family name"), max_length=100)
    address = models.CharField(
        _("address"), max_length=100, blank=True, help_text=_("Mr./Ms./...")
    )
    address_on_first_name_terms = models.BooleanField(
        _("address on first-name terms"), default=False
    )
    salutation = models.CharField(
        _("complete salutation"),
        max_length=100,
        help_text=_("Dear John/Dear Ms Smith"),
    )
    date_of_birth = models.DateField(_("date of birth"), blank=True, null=True)
    notes = models.TextField(_("notes"), blank=True)
    organization = models.ForeignKey(
        Organization,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("organization"),
        related_name="people",
    )
    primary_contact = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("primary contact"),
    )
    groups = models.ManyToManyField(Group, verbose_name=_("groups"), blank=True)
    _fts = models.TextField(editable=False, blank=True)

    objects = PersonQuerySet.as_manager()

    class Meta:
        ordering = ["given_name", "family_name"]
        verbose_name = _("person")
        verbose_name_plural = _("people")

    def __str__(self):
        parts = [self.full_name]
        if self.is_archived:
            parts.append(gettext("(archived)"))
        return " ".join(parts)

    @property
    def full_name(self):
        return " ".join(filter(None, (self.given_name, self.family_name)))

    @property
    def name_with_organization(self):
        if self.organization and not self.organization.is_private_person:
            return "%s / %s" % (self.organization, self.full_name)
        return self.full_name

    def save(self, *args, **kwargs):
        self._fts = " ".join(
            itertools.chain(
                [self.organization.name if self.organization else ""],
                (detail.phone_number for detail in self.phonenumbers.all()),
                (detail.email for detail in self.emailaddresses.all()),
                (detail.postal_address for detail in self.postaladdresses.all()),
            )
        )
        super().save(*args, **kwargs)

    save.alters_data = True


_home_re = re.compile(r"(home|hause|privat)", re.I)


class PersonDetail(Model):
    WEIGHTS = (
        (re.compile(r"mobil", re.I), 30),
        (re.compile(r"(work|arbeit)", re.I), 20),
        (_home_re, 10),
        (re.compile(r"(organization|firm)", re.I), -100),
    )

    type = models.CharField(_("type"), max_length=40, default=_("work"))
    weight = models.SmallIntegerField(_("weight"), default=0, editable=False)

    class Meta:
        abstract = True

    @property
    def urls(self):
        return self.person.urls

    def get_absolute_url(self):
        return self.person.urls["detail"]

    def save(self, *args, **kwargs):
        self.weight = sum(
            (weight for regex, weight in self.WEIGHTS if regex.search(str(self.type))),
            0,
        )
        super().save(*args, **kwargs)

    save.alters_data = True


class PhoneNumber(PersonDetail):
    person = models.ForeignKey(
        Person,
        on_delete=models.CASCADE,
        verbose_name=_("person"),
        related_name="phonenumbers",
    )
    phone_number = models.CharField(_("phone number"), max_length=100)

    class Meta:
        ordering = ("-weight", "id")
        verbose_name = _("phone number")
        verbose_name_plural = _("phone numbers")

    def __str__(self):
        return self.phone_number

    def clean_fields(self, exclude):
        super().clean_fields(exclude)
        errors = {}
        try:
            number = phonenumbers.parse(self.phone_number, "CH")
        except phonenumbers.NumberParseException as exc:
            errors["phone_number"] = str(exc)
        else:
            if phonenumbers.is_valid_number(number):
                self.phone_number = phonenumbers.format_number(
                    number, phonenumbers.PhoneNumberFormat.E164
                )
            else:
                errors["phone_number"] = _("Phone number invalid.")
        raise_if_errors(errors, exclude)

    @property
    def pretty_number(self):
        try:
            return phonenumbers.format_number(
                phonenumbers.parse(self.phone_number),
                phonenumbers.PhoneNumberFormat.INTERNATIONAL,
            )
        except Exception:
            return self.phone_number


class EmailAddress(PersonDetail):
    person = models.ForeignKey(
        Person,
        on_delete=models.CASCADE,
        verbose_name=_("person"),
        related_name="emailaddresses",
    )
    email = models.EmailField(_("email"), max_length=254)

    class Meta:
        ordering = ("-weight", "id")
        verbose_name = _("email address")
        verbose_name_plural = _("email addresses")

    def __str__(self):
        return self.email


class PostalAddress(PersonDetail):
    person = models.ForeignKey(
        Person,
        on_delete=models.CASCADE,
        verbose_name=_("person"),
        related_name="postaladdresses",
    )
    street = models.CharField(_("street"), max_length=100)
    house_number = models.CharField(_("house number"), max_length=20, blank=True)
    address_suffix = models.CharField(_("address suffix"), max_length=100, blank=True)
    postal_code = models.CharField(_("postal code"), max_length=20)
    city = models.CharField(_("city"), max_length=100)
    country = CountryField(_("country"), default="CH")
    postal_address_override = models.TextField(
        _("override"),
        blank=True,
        help_text=_("Completely overrides the postal address if set."),
    )

    class Meta:
        ordering = ("-weight", "id")
        verbose_name = _("postal address")
        verbose_name_plural = _("postal addresses")

    def __str__(self):
        return self.postal_address

    @property
    def postal_address(self):
        if self.postal_address_override:
            return self.postal_address_override
        lines = [
            self.person.organization.name
            if self.person.organization
            and not self.person.organization.is_private_person
            and not _home_re.search(self.type)
            else "",
            self.person.full_name,
            " ".join(filter(None, (self.street, self.house_number))),
            self.address_suffix,
            " ".join(filter(None, (self.postal_code, self.city))),
            self.country.name if self.country.code != "CH" else "",
        ]
        return "\n".join(filter(None, lines))

    def get_maps_url(self):
        parts = [self.street, self.house_number, self.postal_code, self.city]
        if any(parts):
            return "https://www.google.com/maps?{}".format(
                urlencode({"q": " ".join(parts)})
            )
        return None



from django.test import TestCase

from workbench import factories
from workbench.tools.testing import messages
from workbench.tools.validation import in_days


class ReportingTest(TestCase):
    def test_open_items(self):
        """The open items list offers filtering by cutoff date and XLSX exports"""
        invoice = factories.InvoiceFactory.create(
            invoiced_on=in_days(0),
            due_on=in_days(15),
            subtotal=50,
            status=factories.Invoice.SENT,
            third_party_costs=5,  # key data branch
        )
        for i in range(5):
            factories.InvoiceFactory.create(
                customer=invoice.customer,
                contact=invoice.contact,
                owned_by=invoice.owned_by,
                invoiced_on=in_days(0),
                due_on=in_days(15),
                subtotal=50,
                status=factories.Invoice.SENT,
                third_party_costs=5,  # key data branch
            )

        self.client.force_login(factories.UserFactory.create())
        response = self.client.get("/report/open-items-list/?cutoff_date=bla")
        self.assertRedirects(response, "/report/open-items-list/")
        self.assertEqual(messages(response), ["Form was invalid."])
        response = self.client.get("/report/open-items-list/")
        self.assertContains(response, '<th class="text-right">300.00</th>')

        response = self.client.get(
            "/report/open-items-list/?cutoff_date={}".format(in_days(-1).isoformat())
        )
        self.assertContains(response, '<th class="text-right">0.00</th>')

        self.assertEqual(
            self.client.get("/report/open-items-list/?export=xlsx").status_code, 200
        )
        # print(response, response.content.decode("utf-8"))

        # Hit the key data view to cover some branches and verify that it does not crash
        self.assertEqual(self.client.get("/report/key-data/").status_code, 200)



"""
PostgreSQL full text search support

An example migration for adding full text search support to a model would
look like this::

    # -*- coding: utf-8 -*-
    from __future__ import unicode_literals
    from django.db import models, migrations
    from workbench.tools import search

    class Migration(migrations.Migration):
        dependencies = [
            # ...
        ]

        operations = [
            migrations.RunSQL(
                search.create_structure("database_table")
            ),
            migrations.RunSQL(
                search.fts("database_table", ["field1", "field"])
            ),
        ]
"""

import re


def drop_old_shit(table):
    return """\
DROP TRIGGER IF EXISTS {table}_fts_update_trigger ON {table};
DROP TRIGGER IF EXISTS {table}_fts_insert_trigger ON {table};
DROP FUNCTION IF EXISTS {table}_fts_document (integer);
DROP FUNCTION IF EXISTS {table}_fts_document_trigger ();
""".format(
        table=table
    )


def create_structure(table):
    return """\
ALTER TABLE {table} DROP COLUMN IF EXISTS fts_document;
DROP INDEX IF EXISTS {table}_fts_index;

ALTER TABLE {table} ADD COLUMN fts_document tsvector;
CREATE INDEX {table}_fts_index ON {table} USING gin(fts_document);
""".format(
        table=table
    )


def fts(table, fields):
    return """\
CREATE OR REPLACE FUNCTION {table}_fts() RETURNS trigger AS $$
begin
  new.fts_document:=to_tsvector(
    'pg_catalog.german',
    regexp_replace(
      unaccent(concat_ws(' ', {fields})),
      '[^0-9A-Za-z]+',
      ' '
    )
  );
  return new;
end
$$ LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS {table}_fts_trigger ON {table};
CREATE TRIGGER {table}_fts_trigger BEFORE INSERT OR UPDATE
  ON {table} FOR EACH ROW EXECUTE PROCEDURE {table}_fts();
""".format(
        table=table, fields=", ".join("new.{}".format(field) for field in fields)
    )


def process_query(s):
    """
    Converts the user's search string into something suitable for passing to
    to_tsquery.
    """
    # noqa Thanks https://www.fusionbox.com/blog/detail/partial-word-search-with-postgres-full-text-search-in-django/632/
    query = re.sub(r"[^-+@\w]+", " ", s).strip()
    if query:
        query = re.sub(r"\s+", " & ", query)
        # Support prefix search on the last word. A tsquery of 'toda:*' will
        # match against any words that start with 'toda', which is good for
        # search-as-you-type.
        query += ":*"
    return query


def search(queryset, terms):
    return (
        queryset.extra(
            where=[
                "%s.fts_document"
                " @@ to_tsquery('pg_catalog.german', unaccent(%%s))"
                % (queryset.model._meta.db_table,)
            ],
            params=[process_query(terms)],
        )
        if terms
        else queryset
    )



import datetime as dt

from django import forms
from django.conf import settings
from django.contrib import messages
from django.http import Http404
from django.utils.html import format_html, format_html_join
from django.utils.translation import gettext_lazy as _

from workbench.accounts.models import User
from workbench.contacts.models import Organization, Person
from workbench.deals.models import (
    Attribute,
    AttributeGroup,
    Deal,
    DealAttribute,
    Value,
    ValueType,
)
from workbench.tools.formats import Z2
from workbench.tools.forms import Autocomplete, Form, ModelForm, Textarea
from workbench.tools.validation import in_days
from workbench.tools.xlsx import WorkbenchXLSXDocument


class DealSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    s = forms.ChoiceField(
        choices=(
            ("all", _("All states")),
            ("", _("Open")),
            (
                _("Closed"),
                [row for row in Deal.STATUS_CHOICES if not row[0] == Deal.OPEN],
            ),
        ),
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )
    org = forms.ModelChoiceField(
        queryset=Organization.objects.all(),
        required=False,
        widget=Autocomplete(model=Organization),
        label="",
    )
    owned_by = forms.TypedChoiceField(
        coerce=int,
        required=False,
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["owned_by"].choices = User.objects.choices(
            collapse_inactive=True, myself=True
        )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q"))
        self.should_group_deals = False
        if data.get("s") == "":
            queryset = queryset.filter(status=Deal.OPEN).order_by(
                "-probability", "status", "decision_expected_on", "id"
            )
            self.should_group_deals = True
        elif data.get("s") != "all":
            queryset = queryset.filter(status=data.get("s")).order_by(
                "-closed_on", "-pk"
            )
        queryset = self.apply_renamed(queryset, "org", "customer")
        queryset = self.apply_owned_by(queryset)
        return queryset.select_related("owned_by", "customer", "contact__organization")

    def response(self, request, queryset):
        if request.GET.get("export") == "xlsx":
            xlsx = WorkbenchXLSXDocument()
            additional = []
            values = {
                (v.deal_id, v.type_id): v.value
                for v in Value.objects.filter(deal__in=queryset)
            }
            attributes = {
                (a.deal_id, a.attribute.group_id): a.attribute
                for a in DealAttribute.objects.filter(deal__in=queryset).select_related(
                    "attribute"
                )
            }

            for vt in ValueType.objects.all():
                additional.append(
                    (
                        "{}: {}".format(Value._meta.verbose_name, vt),
                        (lambda id: lambda deal: values.get((deal.id, id)))(vt.id),
                    )
                )
            for ag in AttributeGroup.objects.all():
                additional.append(
                    (
                        "{}: {}".format(Attribute._meta.verbose_name, ag),
                        (lambda id: lambda deal: attributes.get((deal.id, id)))(ag.id),
                    )
                )
            xlsx.table_from_queryset(queryset, additional=additional)
            return xlsx.to_response("deals.xlsx")


def warn_if_not_in_preparation(form):
    if form.instance.closed_on and form.request.method == "GET":
        messages.warning(
            form.request,
            _(
                "This deal is already closed."
                " I assume you know what you're doing and wont stand in the way."
            ),
        )


class DealForm(ModelForm):
    user_fields = default_to_current_user = ("owned_by",)

    class Meta:
        model = Deal
        fields = (
            "contact",
            "customer",
            "title",
            "description",
            "probability",
            "decision_expected_on",
            "owned_by",
        )
        widgets = {
            "customer": Autocomplete(model=Organization),
            "contact": Autocomplete(model=Person, params={"only_employees": "on"}),
            "probability": forms.RadioSelect,
            "description": Textarea,
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn_if_not_in_preparation(self)

        field = Value._meta.get_field("value")
        values = {v.type_id: v.value for v in self.instance.values.all()}
        for vt in ValueType.objects.all():
            key = "value_{}".format(vt.id)

            if vt.is_archived and vt.id not in values:
                continue

            self.fields[key] = field.formfield(
                label="%s (%s)"
                % (
                    vt.title,
                    _("Total %(currency)s") % {"currency": settings.WORKBENCH.CURRENCY},
                ),
                required=False,
                initial=values.get(vt.id),
            )

        attributes = (
            {a.group_id: a.id for a in self.instance.attributes.all()}
            if self.instance.id
            else {}
        )
        for group in AttributeGroup.objects.all():
            key = "attribute_{}".format(group.id)
            if group.is_archived and group.id not in attributes:
                continue

            self.fields[key] = forms.ModelChoiceField(
                queryset=group.attributes.all(),
                required=group.is_required,
                label=group.title,
                widget=forms.RadioSelect,
                initial=attributes.get(group.id),
            )
            self.fields[key].choices = [
                (a.id, str(a))
                for a in group.attributes.active(include=attributes.get(group.id))
            ]

        self.fields["decision_expected_on"].help_text = format_html(
            "{} {}",
            _("Expected"),
            format_html_join(
                ", ",
                '<a href="#" data-field-value="{}">{}</a>',
                [
                    (in_days(7).isoformat(), _("in one week")),
                    (in_days(30).isoformat(), _("in one month")),
                    (in_days(60).isoformat(), _("in two months")),
                    (in_days(90).isoformat(), _("in three months")),
                ],
            ),
        )

    def clean(self):
        data = super().clean()
        if (
            self.instance.status == Deal.OPEN
            and data.get("probability") == Deal.HIGH
            and not data.get("decision_expected_on")
        ):
            self.add_error(
                "decision_expected_on",
                _("This field is required when probability is high."),
            )
        return data

    def save(self):
        instance = super().save(commit=False)
        values = {}

        for vt in ValueType.objects.all():
            key = "value_{}".format(vt.id)
            if self.cleaned_data.get(key) is not None:
                values[vt] = self.cleaned_data[key]

        instance.value = sum(values.values(), Z2)
        instance.save(skip_value_calculation=True)

        instance.values.exclude(type__in=values.keys()).delete()
        for vt, value in values.items():
            instance.values.update_or_create(type=vt, defaults={"value": value})

        attributes = []
        for group in AttributeGroup.objects.all():
            key = "attribute_{}".format(group.id)

            if self.cleaned_data.get(key) is not None:
                attributes.append(self.cleaned_data.get(key))

        instance.attributes.set(attributes)

        return instance


class SetStatusForm(ModelForm):
    class Meta:
        model = Deal
        fields = ["status", "closing_type", "closing_notice"]
        widgets = {
            "status": forms.RadioSelect,
            "closing_type": forms.RadioSelect,
            "closing_notice": Textarea,
        }

    def __init__(self, *args, **kwargs):
        instance = kwargs["instance"]
        instance.status = int(kwargs["request"].GET.get("status", instance.status))
        super().__init__(*args, **kwargs)
        self.fields["status"].disabled = True

        if instance.status == Deal.OPEN:
            self.fields.pop("closing_type")
            self.fields.pop("closing_notice")

        elif instance.status == Deal.ACCEPTED:
            self.fields["closing_type"].empty_label = None
            self.fields["closing_type"].label = _("Award of contract")
            self.fields["closing_type"].required = True
            self.fields["closing_type"].queryset = self.fields[
                "closing_type"
            ].queryset.filter(represents_a_win=True)

        elif instance.status == Deal.DECLINED:
            self.fields["closing_type"].empty_label = None
            self.fields["closing_type"].label = _("Reason for losing")
            self.fields["closing_type"].required = True
            self.fields["closing_type"].queryset = self.fields[
                "closing_type"
            ].queryset.filter(represents_a_win=False)

        else:
            raise Http404

        if instance.status != Deal.OPEN:
            related_offers = instance.related_offers.select_related(
                "owned_by", "project"
            )
            if related_offers:
                self.fields["related_offers"] = forms.ModelMultipleChoiceField(
                    queryset=related_offers,
                    label=_("Accept offers")
                    if instance.status == Deal.ACCEPTED
                    else _("Decline offers"),
                    required=False,
                    widget=forms.CheckboxSelectMultiple,
                    initial=[
                        offer.id
                        for offer in related_offers
                        if offer.status < offer.ACCEPTED
                    ],
                )
                self.fields["related_offers"].choices = [
                    (
                        offer.id,
                        format_html("{}<br>{}", offer.__html__(), offer.status_badge),
                    )
                    for offer in related_offers
                ]

    def clean(self):
        data = super().clean()
        if data["status"] != Deal.OPEN and not data.get("closing_type"):
            self.add_error(
                "closing_type", _("This field is required when closing a deal.")
            )
        return data

    def save(self, *args, **kwargs):
        instance = super().save(commit=False)
        assert instance.status in {Deal.OPEN, Deal.ACCEPTED, Deal.DECLINED}

        if instance.status == instance.OPEN:
            instance.closed_on = None
            instance.closing_type = None
            instance.closing_notice = ""
        else:
            instance.closed_on = instance.closed_on or dt.date.today()

            projects = set()
            for offer in self.cleaned_data.get("related_offers", ()):
                offer.status = (
                    offer.ACCEPTED
                    if instance.status == Deal.ACCEPTED
                    else offer.DECLINED
                )
                offer.full_clean()
                offer.save()
                projects.add(offer.project)

            for project in projects:
                project.solely_declined_offers_warning(request=self.request)

        instance.save()
        return instance



from collections import OrderedDict

from django import forms
from django.conf import settings
from django.contrib import messages
from django.forms.models import inlineformset_factory
from django.template.defaultfilters import linebreaksbr
from django.utils.html import format_html, format_html_join
from django.utils.translation import gettext_lazy as _, ngettext

from workbench.contacts.models import (
    EmailAddress,
    Group,
    Organization,
    Person,
    PhoneNumber,
    PostalAddress,
)
from workbench.tools.forms import Autocomplete, Form, ModelForm, Textarea, add_prefix
from workbench.tools.models import ProtectedError, SlowCollector
from workbench.tools.substitute_with import substitute_with
from workbench.tools.vcard import person_to_vcard, render_vcard_response
from workbench.tools.xlsx import WorkbenchXLSXDocument


class OrganizationSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    g = forms.ModelChoiceField(
        queryset=Group.objects.all(),
        required=False,
        empty_label=_("All groups"),
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q")).active()
        return self.apply_renamed(queryset, "g", "groups")


class PersonSearchForm(Form):
    q = forms.CharField(
        required=False,
        widget=forms.TextInput(
            attrs={"class": "form-control", "placeholder": _("Search")}
        ),
        label="",
    )
    g = forms.ModelChoiceField(
        queryset=Group.objects.all(),
        required=False,
        empty_label=_("All groups"),
        widget=forms.Select(attrs={"class": "custom-select"}),
        label="",
    )

    def filter(self, queryset):
        data = self.cleaned_data
        queryset = queryset.search(data.get("q")).active()
        return self.apply_renamed(queryset, "g", "groups").select_related(
            "organization"
        )

    def response(self, request, queryset):
        if request.GET.get("export") == "xlsx":
            xlsx = WorkbenchXLSXDocument()
            xlsx.people(queryset)
            return xlsx.to_response("people.xlsx")

        elif request.GET.get("export") == "vcard":
            return render_vcard_response(
                request,
                "\n".join(
                    person_to_vcard(person).serialize()
                    for person in queryset.prefetch_related(
                        "phonenumbers", "emailaddresses", "postaladdresses"
                    )[: settings.BATCH_MAX_ITEMS]
                ),
            )


class OrganizationForm(ModelForm):
    user_fields = default_to_current_user = ("primary_contact",)

    class Meta:
        model = Organization
        fields = (
            "name",
            "is_private_person",
            "notes",
            "primary_contact",
            "default_billing_address",
            "groups",
            "is_archived",
        )
        widgets = {
            "name": Textarea(),
            "notes": Textarea(),
            "default_billing_address": Textarea(),
            "groups": forms.CheckboxSelectMultiple(),
        }


class OrganizationDeleteForm(ModelForm):
    class Meta:
        model = Organization
        fields = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        collector = SlowCollector(using=self.instance._state.db)
        try:
            collector.collect([self.instance])
        except ProtectedError:
            self.fields["substitute_with"] = forms.ModelChoiceField(
                Organization.objects.exclude(pk=self.instance.pk),
                widget=Autocomplete(model=Organization),
                label=_("substitute with"),
            )
            if self.request.method == "GET":
                messages.info(
                    self.request,
                    _(
                        "This organization has related objects. It is"
                        " therefore not possible to delete it, but you"
                        " may reassign the related objects to a different"
                        " organization."
                        " Alternatively, you can also archive the organization."
                    ),
                )

    def delete(self):
        if "substitute_with" in self.fields:
            substitute_with(
                to_delete=self.instance, keep=self.cleaned_data["substitute_with"]
            )
        else:
            self.instance.delete()


class PersonForm(ModelForm):
    user_fields = default_to_current_user = ["primary_contact"]

    class Meta:
        model = Person
        fields = (
            "address",
            "given_name",
            "family_name",
            "address_on_first_name_terms",
            "salutation",
            "date_of_birth",
            "notes",
            "organization",
            "primary_contact",
            "groups",
            "is_archived",
        )
        widgets = {
            "notes": Textarea(),
            "organization": Autocomplete(model=Organization),
            "groups": forms.CheckboxSelectMultiple(),
        }

    def __init__(self, *args, **kwargs):
        request = kwargs["request"]
        initial = kwargs.setdefault("initial", {})
        if organization := request.GET.get("organization"):
            initial["organization"] = organization

        super().__init__(*args, **kwargs)
        kwargs.pop("request")
        self.formsets = (
            OrderedDict(
                (
                    ("phonenumbers", PhoneNumberFormset(*args, **kwargs)),
                    ("emailaddresses", EmailAddressFormset(*args, **kwargs)),
                    ("postaladdresses", PostalAddressFormset(*args, **kwargs)),
                )
            )
            if self.instance.pk
            else OrderedDict()
        )

    def clean(self):
        data = super().clean()
        if data.get("salutation") and len(data.get("salutation").split()) < 2:
            self.add_warning(
                _("This does not look right. Please add a full salutation."),
                code="short-salutation",
            )
        if self.instance.pk and "organization" in self.changed_data:
            from workbench.invoices.models import Invoice
            from workbench.projects.models import Project

            related = []
            invoices = Invoice.objects.filter(contact=self.instance).count()
            projects = Project.objects.filter(contact=self.instance).count()

            for model, count in [
                (Invoice, invoices),
                (Project, projects),
            ]:
                if count:
                    related.append(
                        "%s %s"
                        % (
                            count,
                            ngettext(
                                model._meta.verbose_name,
                                model._meta.verbose_name_plural,
                                count,
                            ),
                        )
                    )

            if related:
                self.add_warning(
                    _(
                        "This person is the contact of the following related objects:"
                        " %s. Modifying this record may be more confusing than"
                        " archiving this one and creating a new record instead."
                    )
                    % ", ".join(related),
                    code="organization-update",
                )

        return data

    def is_valid(self):
        return all(
            [super().is_valid()]
            + [formset.is_valid() for formset in self.formsets.values()]
        )

    def save(self, commit=True):
        instance = super().save()
        for formset in self.formsets.values():
            formset.save()
        return instance


PhoneNumberFormset = inlineformset_factory(
    Person,
    PhoneNumber,
    fields=("type", "phone_number"),
    extra=0,
    widgets={
        "type": forms.TextInput(
            attrs={"class": "form-control short", "placeholder": _("type")}
        ),
        "phone_number": forms.TextInput(attrs={"class": "form-control"}),
    },
)

EmailAddressFormset = inlineformset_factory(
    Person,
    EmailAddress,
    fields=("type", "email"),
    extra=0,
    widgets={
        "type": forms.TextInput(
            attrs={"class": "form-control short", "placeholder": _("type")}
        ),
        "email": forms.TextInput(attrs={"class": "form-control"}),
    },
)


class PostalAddressForm(forms.ModelForm):
    class Meta:
        model = PostalAddress
        fields = (
            "type",
            "street",
            "house_number",
            "address_suffix",
            "postal_code",
            "city",
            "country",
            "postal_address_override",
        )
        widgets = {
            "type": forms.TextInput(
                attrs={"class": "form-control short", "placeholder": _("type")}
            ),
            "postal_address_override": Textarea(
                attrs={"class": "form-control", "rows": 3}
            ),
        }


PostalAddressFormset = inlineformset_factory(
    Person, PostalAddress, form=PostalAddressForm, extra=0
)


class PostalAddressSelectionForm(ModelForm):
    def add_postal_address_selection_if_empty(
        self, *, organization=None, person=None, for_billing=False
    ):
        postal_addresses = []

        if person:
            if (
                for_billing
                and person.organization
                and person.organization.default_billing_address
            ):
                postal_addresses.append(
                    (
                        _("default billing address"),
                        person.organization.default_billing_address,
                    )
                )
            postal_addresses.extend(
                (pa.type, pa.postal_address)
                for pa in PostalAddress.objects.filter(person=person).select_related(
                    "person__organization"
                )
            )

        if organization and (not person or not postal_addresses):
            if for_billing and organization.default_billing_address:
                postal_addresses.append(
                    (_("default billing address"), organization.default_billing_address)
                )
            postal_addresses.extend(
                (pa.type, pa.postal_address)
                for pa in PostalAddress.objects.filter(
                    person__organization=organization
                )
                .exclude(person=person)
                .select_related("person__organization")
            )

        if postal_addresses:
            self.fields["postal_address"].help_text = format_html(
                '<strong>{}</strong><br><div class="'
                ' list-group list-group-horizontal flex-wrap">{}</div>',
                _("Select one of the following addresses:"),
                format_html_join(
                    "",
                    '<span class="list-group-item list-group-item-action w-auto"'
                    ' data-field-value="{}"><strong>{}</strong><br>{}</span>',
                    [(pa, type, linebreaksbr(pa)) for type, pa in postal_addresses],
                ),
            )
            # repr(postal_addresses)

    def clean(self):
        data = super().clean()

        postal_address = data.get("postal_address", "")
        if len(postal_address.strip().splitlines()) < 3:
            self.add_warning(
                _(
                    "The postal address should probably be at least three"
                    " lines long."
                ),
                code="short-postal-address",
            )

        return data


@add_prefix("modal")
class PersonAutocompleteForm(forms.Form):
    person = forms.ModelChoiceField(
        queryset=Person.objects.all(), widget=Autocomplete(model=Person), label=""
    )



import datetime as dt
from copy import deepcopy
from decimal import Decimal as D
from itertools import chain

from django.conf import settings
from django.utils.text import Truncator, capfirst
from django.utils.translation import activate, gettext as _

from pdfdocument.document import (
    TA_RIGHT,
    Frame,
    MarkupParagraph,
    NextPageTemplate,
    PageTemplate,
    PDFDocument as _PDFDocument,
    cm,
    colors,
    getSampleStyleSheet,
    mm,
    register_fonts_from_paths,
    sanitize,
)
from pdfdocument.utils import pdf_response as _pdf_response

from workbench.tools.formats import currency, hours, local_date_format
from workbench.tools.models import ModelWithTotal


register_fonts_from_paths(font_name="Rep", **settings.WORKBENCH.FONTS)


Z = D("0.00")


class Empty(object):
    pass


def style(base, **kwargs):
    style = deepcopy(base)
    for key, value in kwargs.items():
        setattr(style, key, value)
    return style


class PDFDocument(_PDFDocument):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("font_name", "Rep")
        kwargs.setdefault("font_size", 8.5)
        super().__init__(*args, **kwargs)

    def generate_style(self, *args, **kwargs):
        self.style = Empty()
        self.style.fontName = self.font_name
        self.style.fontSize = self.font_size

        self.style.normal = style(
            getSampleStyleSheet()["Normal"],
            fontName="Rep",
            fontSize=self.style.fontSize,
            firstLineIndent=0,
        )
        self.style.normalWithExtraLeading = style(
            self.style.normal,
            leading=1.75 * self.style.fontSize,
        )
        self.style.heading1 = style(
            self.style.normal,
            fontName="Rep-Bold",
            fontSize=1.25 * self.style.fontSize,
            leading=1.25 * self.style.fontSize,
        )
        self.style.heading2 = style(
            self.style.normal,
            fontName="Rep-Bold",
            fontSize=1.15 * self.style.fontSize,
            leading=1.15 * self.style.fontSize,
        )
        self.style.heading3 = style(
            self.style.normal,
            fontName="Rep-Bold",
            fontSize=1.1 * self.style.fontSize,
            leading=1.2 * self.style.fontSize,
        )

        self.style.right = style(self.style.normal, alignment=TA_RIGHT)

        self.style.small = style(self.style.normal, fontSize=self.style.fontSize * 0.9)
        self.style.smaller = style(
            self.style.normal, fontSize=self.style.fontSize * 0.75
        )
        self.style.bold = style(self.style.normal, fontName="Rep-Bold")
        self.style.paragraph = style(self.style.normal, spaceBefore=1, spaceAfter=1)
        self.style.table = (
            ("FONT", (0, 0), (-1, -1), "Rep", self.style.fontSize),
            ("TOPPADDING", (0, 0), (-1, -1), 0),
            ("BOTTOMPADDING", (0, 0), (-1, -1), 1),
            ("LEFTPADDING", (0, 0), (-1, -1), 0),
            ("RIGHTPADDING", (0, 0), (-1, -1), 0),
            ("FIRSTLINEINDENT", (0, 0), (-1, -1), 0),
            ("ALIGN", (0, 0), (-1, -1), "LEFT"),
            ("VALIGN", (0, 0), (-1, -1), "TOP"),
        )

        self.style.tableHeadLine = self.style.table + (
            ("ALIGN", (1, 0), (-1, -1), "RIGHT"),
            ("RIGHTPADDING", (0, 0), (0, -1), 2 * mm),
            ("LINEABOVE", (0, 0), (-1, 0), 0.2, colors.black),
            ("LINEBELOW", (0, 0), (-1, 0), 0.2, colors.black),
            ("FONT", (0, 0), (-1, 0), "Rep-Bold", self.style.fontSize),
            ("TOPPADDING", (0, 0), (-1, 0), 1),
            ("BOTTOMPADDING", (0, 0), (-1, 0), 2),
        )

        self.style.tableHead = self.style.tableHeadLine + (
            ("TOPPADDING", (0, 1), (-1, 1), 1),
        )

        self.bounds = Empty()
        self.bounds.N = 275 * mm
        self.bounds.E = 190 * mm
        self.bounds.S = 18 * mm
        self.bounds.W = 20 * mm
        self.bounds.outsideN = self.bounds.N + 5 * mm
        self.bounds.outsideS = self.bounds.S - 5 * mm
        self.style.tableColumns = (self.bounds.E - self.bounds.W - 32 * mm, 32 * mm)
        self.style.tableColumnsLeft = list(reversed(self.style.tableColumns))
        self.style.tableThreeColumns = (
            self.bounds.E - self.bounds.W - 32 * mm,
            16 * mm,
            16 * mm,
        )

        frame_kwargs = {
            "showBoundary": self.show_boundaries,
            "leftPadding": 0,
            "rightPadding": 0,
            "topPadding": 0,
            "bottomPadding": 0,
        }

        self.address_frame = Frame(
            self.bounds.W,
            20.2 * cm,
            self.bounds.E - self.bounds.W,
            40 * mm,
            **frame_kwargs
        )
        self.rest_frame = Frame(
            self.bounds.W,
            self.bounds.S,
            self.bounds.E - self.bounds.W,
            18.2 * cm,
            **frame_kwargs
        )
        self.full_frame = Frame(
            self.bounds.W,
            self.bounds.S,
            self.bounds.E - self.bounds.W,
            self.bounds.N - self.bounds.S,
            **frame_kwargs
        )

    def init_letter(self, *, page_fn=None, page_fn_later=None):
        page_fn = page_fn or self.stationery()
        self.generate_style()
        self.doc.addPageTemplates(
            [
                PageTemplate(
                    id="First",
                    frames=[self.address_frame, self.rest_frame],
                    onPage=page_fn,
                ),
                PageTemplate(
                    id="Later",
                    frames=[self.full_frame],
                    onPage=page_fn_later or page_fn,
                ),
            ]
        )
        self.story.append(NextPageTemplate("Later"))

    def init_report(self, *, page_fn=None):
        page_fn = page_fn or self.stationery()
        self.generate_style()
        self.doc.addPageTemplates(
            [PageTemplate(id="Page", frames=[self.full_frame], onPage=page_fn)]
        )

    def stationery(self):
        pdf = self

        def _fn(canvas, doc):
            canvas.saveState()

            canvas.setFont(pdf.style.fontName + "-Bold", 10)
            canvas.drawString(
                pdf.bounds.W, pdf.bounds.outsideN, settings.WORKBENCH.PDF_COMPANY
            )

            canvas.setFont(pdf.style.fontName, 6)
            canvas.drawString(
                pdf.bounds.W, pdf.bounds.outsideS, settings.WORKBENCH.PDF_ADDRESS
            )

            canvas.setFont(pdf.style.fontName, 6)
            canvas.drawRightString(
                pdf.bounds.E,
                pdf.bounds.outsideS,
                _("Page %d")
                % (
                    doc.page - doc.restartDocPageNumbers[doc.restartDocIndex - 1]
                    if doc.restartDocIndex
                    else doc.page
                ),
            )

            canvas.restoreState()

            pdf.draw_watermark(canvas)

        return _fn

    def postal_address(self, postal_address):
        self.p(postal_address)
        self.next_frame()

    def services_row(self, service):
        is_optional = getattr(service, "is_optional", False)
        return [
            (
                MarkupParagraph(
                    "<b>%s</b> %s<br/>%s"
                    % (
                        sanitize(service.title),
                        _("(optional)") if is_optional else "",
                        sanitize(service.description),
                    ),
                    self.style.normal,
                ),
                MarkupParagraph(
                    "<i>%s</i>" % currency(service.service_cost.quantize(Z)),
                    self.style.right,
                )
                if is_optional
                else "",
                "" if is_optional else currency(service.service_cost.quantize(Z)),
            )
        ]

    def services_row_with_details(self, service):
        is_optional = getattr(service, "is_optional", False)
        return [
            (
                MarkupParagraph(
                    "<b>%s</b> %s<br/>%s"
                    % (
                        sanitize(service.title),
                        _("(optional)") if is_optional else "",
                        sanitize(service.description),
                    ),
                    self.style.normal,
                ),
                "",
                "",
            ),
            (
                MarkupParagraph(
                    ", ".join(
                        filter(
                            None,
                            [
                                (
                                    "%s %s à %s/h"
                                    % (
                                        hours(service.effort_hours),
                                        service.effort_type,
                                        currency(service.effort_rate),
                                    )
                                )
                                if service.effort_hours and service.effort_rate
                                else "",
                                ("%s %s" % (currency(service.cost), _("fixed costs")))
                                if service.cost
                                else "",
                            ],
                        )
                    ),
                    self.style.normalWithExtraLeading,
                ),
                MarkupParagraph(
                    "<i>%s</i>" % currency(service.service_cost.quantize(Z)),
                    self.style.right,
                )
                if is_optional
                else "",
                "" if is_optional else currency(service.service_cost.quantize(Z)),
            ),
        ]

    def table_services(self, services, *, show_details=False):
        if not services:
            return

        fn = self.services_row_with_details if show_details else self.services_row
        self.table(
            [(_("Services"), "", "")]
            + list(chain.from_iterable(fn(service) for service in services)),
            self.style.tableThreeColumns,
            self.style.tableHead,
        )

    def table_total(self, instance):
        total = [(_("subtotal"), currency(instance.subtotal.quantize(Z)))]
        if instance.discount:
            total.append((_("discount"), currency(-instance.discount.quantize(Z))))
        if getattr(instance, "down_payment_total", None):
            for invoice in instance.down_payment_invoices.all():
                total.append(
                    (
                        MarkupParagraph(
                            "%s: %s (%s)"
                            % (
                                _("Down payment"),
                                Truncator(invoice).chars(60),
                                invoice.invoiced_on.strftime("%d.%m.%Y")
                                if invoice.invoiced_on
                                else _("NO DATE YET"),
                            ),
                            self.style.normal,
                        ),
                        currency(-invoice.total_excl_tax.quantize(Z)),
                    )
                )
        if instance.tax_amount:
            total.append(
                (
                    "%0.1f%% %s" % (instance.tax_rate, _("tax")),
                    currency(instance.tax_amount.quantize(Z)),
                )
            )

        if len(total) > 1:
            self.table(total, self.style.tableColumns, self.style.tableHead)
            self.spacer(0.7 * mm)

        self.table(
            [(instance.total_title, currency(instance.total.quantize(Z)))],
            self.style.tableColumns,
            self.style.tableHeadLine,
        )

    def process_services_letter(self, instance, *, watermark, details, footer):
        self.watermark(watermark)
        self.postal_address(instance.postal_address)
        self.h1(instance.title)
        self.spacer(2 * mm)
        self.table(details, self.style.tableColumnsLeft, self.style.table)
        if instance.description:
            self.spacer(5 * mm)
            self.p(instance.description)
        self.spacer(2 * mm)
        if getattr(instance, "service_period", None):
            self.p("%s: %s" % (_("Service period"), instance.service_period))
            self.spacer()
        self.table_services(
            instance.services.all(),
            show_details=getattr(instance, "show_service_details", False),
        )
        self.table_total(instance)
        self.spacer(2 * mm)
        self.p(footer)

    def process_offer(self, offer):
        self.process_services_letter(
            offer,
            watermark=""
            if offer.status in {offer.OFFERED, offer.ACCEPTED}
            else str(offer.get_status_display()),
            details=[
                (_("offer"), offer.code),
                (
                    _("date"),
                    (
                        local_date_format(offer.offered_on)
                        if offer.offered_on
                        else MarkupParagraph(
                            "<b>%s</b>" % _("NO DATE YET"), style=self.style.bold
                        )
                    ),
                ),
                (_("Our reference"), offer.owned_by.get_full_name()),
                (
                    capfirst(_("valid until")),
                    (
                        local_date_format(offer.valid_until)
                        if offer.valid_until
                        else MarkupParagraph(
                            "<b>%s</b>" % _("NO DATE YET"), style=self.style.bold
                        )
                    ),
                ),
            ],
            footer=settings.WORKBENCH.PDF_OFFER_TERMS,
        )

    def process_invoice(self, invoice):
        self.process_services_letter(
            invoice,
            watermark=""
            if invoice.status in {invoice.SENT, invoice.PAID}
            else str(invoice.get_status_display()),
            details=[
                (
                    _("down payment invoice")
                    if invoice.type == invoice.DOWN_PAYMENT
                    else _("invoice"),
                    invoice.code,
                ),
                (
                    _("date"),
                    (
                        local_date_format(invoice.invoiced_on)
                        if invoice.invoiced_on
                        else MarkupParagraph(
                            "<b>%s</b>" % _("NO DATE YET"), style=self.style.bold
                        )
                    ),
                ),
                (_("Our reference"), invoice.owned_by.get_full_name()),
                ("MwSt.-Nr.", settings.WORKBENCH.PDF_VAT_NO),
            ],
            footer=settings.WORKBENCH.PDF_INVOICE_PAYMENT
            % {
                "code": invoice.code,
                "due": (
                    local_date_format(invoice.due_on)
                    if invoice.due_on
                    else _("NO DATE YET")
                ),
            },
        )

    def offers_pdf(self, *, project, offers):
        self.init_letter()
        self.p(offers[-1].postal_address)
        self.next_frame()
        self.p("Zürich, %s" % local_date_format(dt.date.today()))
        self.spacer()
        self.h1(project.title)
        if project.description:
            self.spacer(2 * mm)
            self.p(project.description)
        self.spacer()
        self.table(
            [
                tuple(
                    capfirst(title)
                    for title in (_("offer"), _("offered on"), _("total"))
                )
            ]
            + [
                (
                    MarkupParagraph(offer.title, self.style.normal),
                    local_date_format(offer.offered_on)
                    if offer.offered_on
                    else MarkupParagraph(
                        "<b>%s</b>" % _("NO DATE YET"), style=self.style.bold
                    ),
                    currency(offer.total_excl_tax),
                )
                for offer in offers
            ],
            (self.bounds.E - self.bounds.W - 40 * mm, 24 * mm, 16 * mm),
            self.style.tableHead + (("ALIGN", (1, 0), (1, -1), "LEFT"),),
        )

        total = ModelWithTotal(
            subtotal=sum((offer.total_excl_tax for offer in offers), Z),
            discount=Z,
            liable_to_vat=offers[0].liable_to_vat,
            tax_rate=offers[0].tax_rate,
        )
        total._calculate_total()
        total.total_title = offers[0].total_title
        self.table_total(total)

        self.restart()
        for offer in offers:
            self.init_letter()
            self.process_offer(offer)
            self.restart()

        self.generate()

    def dunning_letter(self, *, invoices):
        self.init_letter()
        self.p(invoices[-1].postal_address)
        self.next_frame()
        self.p("Zürich, %s" % local_date_format(dt.date.today()))
        self.spacer()
        self.h1("Zahlungserinnerung")
        self.spacer()
        self.mini_html(
            (
                """\
<p>Sehr geehrte Damen und Herren</p>
<p>Bei der beiliegenden Rechnung konnten wir leider noch keinen Zahlungseingang
verzeichnen. Wir bitten Sie, den ausstehenden Betrag innerhalb von 10 Tagen auf
das angegebene Konto zu überweisen. Bei allfälligen Unstimmigkeiten setzen Sie
sich bitte mit uns in Verbindung.</p>
<p>Falls sich Ihre Zahlung mit diesem Schreiben kreuzt, bitten wir Sie, dieses
als gegenstandslos zu betrachten.</p>
<p>Freundliche Grüsse</p>
<p>%s</p>"""
                if len(invoices) == 1
                else """\
<p>Sehr geehrte Damen und Herren</p>
<p>Bei den beiliegenden Rechnungen konnten wir leider noch keinen Zahlungseingang
verzeichnen. Wir bitten Sie, die ausstehenden Beträge innerhalb von 10 Tagen auf
das angegebene Konto zu überweisen. Bei allfälligen Unstimmigkeiten setzen Sie
sich bitte mit uns in Verbindung.</p>
<p>Falls sich Ihre Zahlung mit diesem Schreiben kreuzt, bitten wir Sie, dieses
als gegenstandslos zu betrachten.</p>
<p>Freundliche Grüsse</p>
<p>%s</p>"""
            )
            % settings.WORKBENCH.PDF_COMPANY
        )
        self.spacer()
        self.table(
            [tuple(capfirst(title) for title in (_("invoice"), _("date"), _("total")))]
            + [
                (
                    MarkupParagraph(invoice.title, self.style.normal),
                    local_date_format(invoice.invoiced_on),
                    currency(invoice.total),
                )
                for invoice in invoices
            ],
            (self.bounds.E - self.bounds.W - 40 * mm, 24 * mm, 16 * mm),
            self.style.tableHead + (("ALIGN", (1, 0), (1, -1), "LEFT"),),
        )
        self.restart()
        for invoice in invoices:
            self.init_letter(page_fn=self.stationery())
            self.process_invoice(invoice)
            self.restart()

    def table_columns(self, columns):
        given = sum(filter(None, columns))
        calculated = (self.bounds.E - self.bounds.W - given) / columns.count(None)
        return [calculated if col is None else col for col in columns]


def pdf_response(*args, **kwargs):
    activate(settings.WORKBENCH.PDF_LANGUAGE)
    kwargs["pdfdocument"] = PDFDocument
    return _pdf_response(*args, **kwargs)



# Generated by Django 3.0.5 on 2020-04-23 09:22

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0018_auto_20200402_2212"),
        ("timer", "0004_timestamp_logged_break"),
    ]

    operations = [
        migrations.AddField(
            model_name="timestamp",
            name="project",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="projects.Project",
                verbose_name="project",
            ),
        ),
    ]



from django.core.management import BaseCommand
from django.utils.translation import activate

from workbench.accounts.middleware import set_user_name
from workbench.circles.tasks import update_circles


class Command(BaseCommand):
    help = "Update circles"

    def handle(self, **options):
        activate("de")
        set_user_name("GlassFrog")
        update_circles()



# Generated by Django 3.0.2 on 2020-01-30 13:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0019_auto_20191002_1005"),
    ]

    operations = [
        migrations.AddField(
            model_name="recurringinvoice",
            name="create_invoice_on_day",
            field=models.IntegerField(
                default=-20,
                help_text="Invoices are created 20 days before their period begins by default.",
                verbose_name="create invoice on day",
            ),
        ),
    ]



import datetime as dt

from workbench.invoices.utils import recurring
from workbench.reporting.models import Accruals


def create_accruals_for_last_month():
    today = dt.date.today()
    start = today.replace(year=today.year - 2, day=1)

    for day in recurring(start, "monthly"):
        if day > today:
            break
        Accruals.objects.for_cutoff_date(day - dt.timedelta(days=1))



from workbench.accounts.models import User


class AuthBackend:
    def get_user(self, user_id):
        return User.objects.filter(pk=user_id).first()

    def authenticate(self, request, **kwargs):
        # Fine since email is unique
        return User.objects.filter(email=kwargs.get("email")).first()



from django.contrib import messages
from django.http import HttpResponseRedirect
from django.views.decorators.http import require_POST

from workbench.notes.forms import NoteForm


@require_POST
def add_note(request):
    form = NoteForm(request.POST, request=request)
    if form.is_valid():
        form.save()
    else:
        for msg in form.non_field_errors():
            messages.error(request, msg)
    return HttpResponseRedirect(request.POST.get("next") or "/")
